from urllib.parse import urlparse

import redis
from celery import current_app


def celery_redis_instance():
    broker_uri = current_app.connection().as_uri(include_password=True)
    url = urlparse(broker_uri)
    return redis.Redis(host=url.hostname, port=url.port, password=url.password)



from typing import Optional, List

from .. import current_context as context
from ..util.events import create_event


def notify(notification_type, data=None, task_id=None, root_id=None):
    task_id = task_id or context.current.task_id
    root_id = root_id or context.current.root_id

    create_event(
        'notify',
        root_id=root_id,
        task_id=task_id,
        data={
            'type': notification_type,
            'data': data or {}
        },
        binary=None
    )


def workflow_complete(task_id: str, root_id: str, status: str):
    notify(
        notification_type='workflow_complete',
        data=dict(status=status),
        task_id=task_id,
        root_id=root_id
    )


def send_email(
        subject: str,
        body: str,
        mime_type: Optional[str] = None,
        sender: Optional[str] = None,
        receivers: Optional[List[str]] = None
):
    if isinstance(receivers, str):
        receivers = [receivers]

    notify(
        notification_type='custom-email',
        data={
            'subject': subject,
            'mime_type': mime_type,
            'body': body,
            'from': sender,
            'to': receivers or []
        }
    )



from typing import List

from sqlalchemy import desc

from riberry import model, policy


def profile() -> model.auth.User:
    return policy.context.subject


def latest_notifications():
    user = policy.context.subject
    return model.misc.UserNotification.query().filter_by(user=user).order_by(
        desc(model.misc.UserNotification.created)
    ).limit(32).all()


def unread_notification_count():
    user = policy.context.subject
    return model.misc.UserNotification.query().filter_by(read=False, user=user).count()


def mark_notifications_as_read(notification_ids: List):
    user = policy.context.subject
    notifications: List[model.misc.UserNotification] = model.misc.UserNotification.query().filter(
        (model.misc.UserNotification.id.in_(notification_ids)) &
        (model.misc.UserNotification.user == user) &
        (model.misc.UserNotification.read == False)
    ).all()

    for notification in notifications:
        notification.read = True


def mark_all_notifications_as_read():
    user = policy.context.subject
    notifications: List[model.misc.UserNotification] = model.misc.UserNotification.query().filter_by(
        user=user,
        read=False
    ).all()

    for notification in notifications:
        notification.read = True




from typing import List

from riberry import model, policy


@policy.context.post_filter(action='view')
def all_applications() -> List[model.application.Application]:
    return model.application.Application.query().all()


@policy.context.post_authorize(action='view')
def application_by_id(application_id) -> model.application.Application:
    return model.application.Application.query().filter_by(id=application_id).one()


@policy.context.post_authorize(action='view')
def application_by_internal_name(internal_name) -> model.application.Application:
    return model.application.Application.query().filter_by(internal_name=internal_name).one()


def create_application(name, internal_name, description, type, document):
    app = model.application.Application(
        name=name,
        internal_name=internal_name,
        description=description,
        type=type,
        document=model.misc.Document(content=document) if document else None
    )

    policy.context.authorize(app, action='create')
    model.conn.add(app)
    return app


def update_application(application, attributes):
    for attr in {'name', 'description', 'type'} & set(attributes):
        setattr(application, attr, attributes[attr])

    return application



import enum
import json
from datetime import datetime
from typing import List, Optional

from sqlalchemy import Binary, String, Column, Float, ForeignKey, Boolean, DateTime, Index, Enum, UniqueConstraint, sql
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import relationship

from riberry import model
from riberry.model import base


class ResourceType(enum.Enum):
    form = 'Form'
    user = 'User'
    job = 'Job'
    job_execution = 'JobExecution'

    def __repr__(self):
        return repr(self.value)


class Document(base.Base):
    __tablename__ = 'document'
    __reprattrs__ = ['type']

    id = base.id_builder.build()
    type: str = Column(String(24), nullable=False, default='markdown')
    content: bytes = Column(Binary, nullable=False)


class Event(base.Base):
    __tablename__ = 'event'
    __reprattrs__ = ['name', 'root_id']

    # columns
    id = base.id_builder.build()
    name: str = Column(String(64), nullable=False)
    time: float = Column(Float, nullable=False)
    root_id: str = Column(String(36), nullable=False)
    task_id: str = Column(String(36), nullable=False)
    data: str = Column(String(1024))
    binary: bytes = Column(Binary)


class NotificationType(enum.Enum):
    info = 'info'
    warning = 'warning'
    success = 'success'
    error = 'error'
    alert = 'alert'


class UserNotification(base.Base):
    __tablename__ = 'notification_user'
    __table_args__ = (
        Index('u_n__idx_job_id', 'user_id', 'read'),
    )

    # columns
    id = base.id_builder.build()
    user_id = Column(base.id_builder.type, ForeignKey('users.id'), nullable=False)
    notification_id = Column(base.id_builder.type, ForeignKey('notification.id'), nullable=False)
    created: datetime = Column(DateTime(timezone=True), default=base.utc_now)
    read = Column(Boolean, nullable=False, default=False)

    # associations
    user: 'model.auth.User' = relationship('User', back_populates='notifications')
    notification: 'Notification' = relationship('Notification', back_populates='user_notifications')


class Notification(base.Base):
    __tablename__ = 'notification'

    # columns
    id = base.id_builder.build()
    type = Column(Enum(NotificationType), nullable=False, default=NotificationType.info)
    message = Column(String(128), nullable=False)

    # associations
    user_notifications: List['UserNotification'] = relationship('UserNotification', back_populates='notification')
    targets: List['NotificationTarget'] = relationship('NotificationTarget', back_populates='notification')


class NotificationTarget(base.Base):
    __tablename__ = 'notification_target'

    # columns
    id = base.id_builder.build()
    notification_id = Column(base.id_builder.type, ForeignKey('notification.id'), nullable=False)
    target = Column(String(128), nullable=False)
    target_id = Column(String(128), nullable=False)
    action = Column(String(32))

    # associations
    notification: 'Notification' = relationship('Notification', back_populates='targets')


class MenuItem(base.Base):
    __tablename__ = 'menu_item'

    # columns
    id = base.id_builder.build()
    parent_id = Column(base.id_builder.type, ForeignKey('menu_item.id'))
    menu_type = Column(String(128), nullable=False)
    type = Column(String(128), nullable=False)
    key = Column(String(128), nullable=False)
    label: Optional[str] = Column(String(128), nullable=True)

    # associations
    parent: 'MenuItem' = relationship('MenuItem', back_populates='children', remote_side=[id])
    children: List['MenuItem'] = relationship('MenuItem', back_populates='parent')


class ResourceData(base.Base):
    __tablename__ = 'resource_data'
    __reprattrs__ = ['name']
    __table_args__ = (
        UniqueConstraint('resource_id', 'resource_type', 'name', name='r_f__id_type_name'),
    )

    # columns
    id = base.id_builder.build()
    resource_id = Column(base.id_builder.type, nullable=True)
    resource_type = Column(Enum(ResourceType), nullable=False)
    name: str = Column(String(256))
    raw_value: bytes = Column('value', Binary, nullable=True)
    lock: str = Column(String(72), nullable=True)
    expiry: datetime = Column(DateTime(timezone=True), nullable=True)
    marked_for_refresh: bool = Column(Boolean, nullable=False, default=False)

    @hybrid_property
    def value(self):
        return json.loads(self.raw_value.decode()) if self.raw_value else None

    @value.setter
    def value(self, value):
        self.raw_value = json.dumps(value).encode()

    @classmethod
    def make_relationship(cls, resource_id, resource_type):
        return relationship(
            'ResourceData',
            primaryjoin=lambda: sql.and_(
                resource_id == ResourceData.resource_id,
                ResourceData.resource_type == resource_type
            ),
            order_by=lambda: ResourceData.id.asc(),
            foreign_keys=lambda: ResourceData.resource_id,
            cascade='save-update, merge, delete, delete-orphan',
        )



from contextlib import contextmanager

import riberry


class Flow:

    def __init__(self, context):
        self.context: riberry.app.context.Context = context
        self.scoped_stream = None
        self.scoped_category = None

    @contextmanager
    def stream_scope(self, stream=None, category=None):
        try:
            self.scoped_stream = stream
            self.scoped_category = category
            yield self
        finally:
            self.scoped_stream = None
            self.scoped_category = None

    def start(self, task, stream: str = None):
        return TaskWrap(
            task,
            __rib_stream_start=True,
            __rib_stream=self.cleanse_stream_name(stream),
        )

    def step(self, task, step: str = None, stream: str = None):
        return TaskWrap(
            task,
            __rib_step=step,
            __rib_stream=self.cleanse_stream_name(stream),
        )

    def end(self, task, stream: str = None):
        return TaskWrap(
            task,
            __rib_stream_end=True,
            __rib_stream=self.cleanse_stream_name(stream),
        )

    def cleanse_stream_name(self, stream: str):
        stream = stream or self.scoped_stream
        if not stream:
            raise ValueError('Stream name cannot be blank')
        return str(stream)


class TaskWrap:

    def __init__(self, func, **kwargs):
        self.func = func
        self.kwargs = kwargs

    @property
    def name(self):
        return self.func.name

    def _mixin_kw(self, kwargs):
        return {**self.kwargs, **kwargs}

    def s(self, *args, **kwargs):
        return self.func.s(*args, **self._mixin_kw(kwargs=kwargs))

    def si(self, *args, **kwargs):
        return self.func.si(*args, **self._mixin_kw(kwargs=kwargs))

    def delay(self, *args, **kwargs):
        return self.func.delay(*args, **self._mixin_kw(kwargs=kwargs))



from contextlib import contextmanager

import riberry
from .artifact import Artifact
from .current import ContextCurrent
from .event_registry import EventRegistry, EventRegistryHelper
from .external_task import ExternalTask
from .flow import Flow
from .input_mapping import InputMappings
from .report import Report
from .shared_data import SharedExecutionData


class Context:

    def __init__(self):
        self.current = ContextCurrent(context=self)
        self.input = InputMappings(context=self)
        self.data = SharedExecutionData(context=self)
        self.flow = Flow(context=self)
        self.artifact = Artifact()
        self.report = Report(context=self)
        self.external_task = ExternalTask(context=self)
        self.event_registry = EventRegistry(context=self)
        self.on = EventRegistryHelper(context=self)

    @contextmanager
    def scope(self, root_id, task_id, stream, category, step):
        with self.current.scope(
            root_id=root_id,
            task_id=task_id,
            stream=stream,
            category=category,
            step=step,
        ):
            yield

    def spawn(self, form_name, job_name=None, input_values=None, input_files=None, owner=None, execute=True):
        return riberry.app.actions.jobs.create_job(
            form_name=form_name,
            job_name=job_name,
            input_files=input_files,
            input_values=input_values,
            owner=owner,
            execute=execute,
        )



from urllib.parse import urlparse

import redis
from celery import current_app, signals
from celery.utils.log import logger
from celery.worker import control

from riberry.celery import client


class ConcurrencyScale:
    all_apps = {}

    def __init__(self, app, target_queues=None):
        self.all_apps[app.main] = self
        self.target_queues = set(target_queues or [])
        self.initial_queues = set()
        self.current_concurrency = None

    @property
    def target_worker_queues(self):
        return self.initial_queues & self.target_queues

    def scale_down(self, consumer):
        current_concurrency = consumer.pool.num_processes

        queues_to_skip = set()
        for queue in consumer.task_consumer.queues:
            self.initial_queues.add(queue.name)
            if queue.name not in self.target_queues:
                queues_to_skip.add(queue.name)
            else:
                logger.info(f'scale-down - removing queue: {queue.name}')
                consumer.cancel_task_queue(queue)

        if self.current_concurrency is None:
            self.current_concurrency = current_concurrency

        if not queues_to_skip and current_concurrency > 0:
            self.current_concurrency = current_concurrency
            logger.info(f'scale-down - shrink amount: {current_concurrency}')
            consumer.pool.shrink(current_concurrency)

    def scale_up(self, consumer):
        current_concurrency = consumer.pool.num_processes

        if self.current_concurrency is None:
            return

        queue_names = {q.name for q in consumer.task_consumer.queues}
        for queue in self.initial_queues:
            if queue not in queue_names:
                logger.info(f'scale-up - adding queue: {queue}')
                consumer.add_task_queue(queue)

        processes_to_add = self.current_concurrency - current_concurrency
        max_processes_to_add = min(processes_to_add, 8)
        if processes_to_add > 0:
            logger.info(f'scale-up - grow amount: {max_processes_to_add}')
            consumer.pool.grow(processes_to_add)

    def scale_to(self, consumer, concurrency):

        queue_names = {q.name for q in consumer.task_consumer.queues}
        if not set(queue_names) & self.target_queues:
            logger.debug(
                f'scale-to: no target queues '
                f'current={", ".join(queue_names or ["N/A"])} target={", ".join(self.target_queues or ["N/A"])}')
            return

        current_concurrency = consumer.pool.num_processes
        process_diff = concurrency - current_concurrency
        process_diff = min(process_diff, 8)
        if process_diff > 0:
            logger.info(f'scale-to: +{process_diff} -> new: {current_concurrency + process_diff} target: {concurrency}')
            consumer.pool.grow(process_diff)
        elif process_diff < 0 and concurrency:
            logger.info(f'scale-to: {process_diff} -> new: {current_concurrency + process_diff} target: {concurrency}')

            consumer.qos.decrement_eventually(consumer.qos.value - 1)
            consumer.qos.update()

            try:
                consumer.pool.shrink(abs(process_diff))
            except Exception as exc:
                logger.error(exc)
                return
        else:
            prefetch_count = (consumer.pool.num_processes * consumer.prefetch_multiplier) - consumer.qos.value
            consumer.qos.increment_eventually(n=prefetch_count)

        self.current_concurrency = process_diff + current_concurrency

    @classmethod
    def instance(cls) -> 'ConcurrencyScale':
        return cls.all_apps.get(current_app.main)


@signals.worker_ready.connect
def scale_down_on_startup(sender, **_):
    scale = ConcurrencyScale.instance()
    scale and scale.scale_down(consumer=sender)


@control.control_command()
def scale_down(state, instance):
    scale = ConcurrencyScale.instance()
    scale and instance and client.is_current_instance(instance_name=instance) and scale.scale_down(
        consumer=state.consumer)


@control.control_command()
def scale_up(state, instance):
    scale = ConcurrencyScale.instance()
    scale and instance and client.is_current_instance(instance_name=instance) and scale.scale_up(
        consumer=state.consumer)


@control.control_command()
def scale_to(state, concurrency, instance):
    scale = ConcurrencyScale.instance()
    scale and instance and client.is_current_instance(instance_name=instance) and scale.scale_to(
        consumer=state.consumer,
        concurrency=concurrency)


@control.control_command()
def worker_task_count(state, instance):
    if not instance or not client.is_current_instance(instance_name=instance):
        return 0

    scale = ConcurrencyScale.instance()
    if not scale or not scale.target_worker_queues:
        return 0

    return sum([len(control.scheduled(state)), len(control.active(state)), len(control.reserved(state))])


def redis_queues_empty_workers_idle(queues):
    broker_uri = current_app.connection().as_uri(include_password=True)
    url = urlparse(broker_uri)
    r = redis.Redis(host=url.hostname, port=url.port, password=url.password)

    separator = '\x06\x16'
    priority_steps = [0, 3, 6, 9]
    for queue in queues:
        for prio in priority_steps:
            queue = f'{queue}{separator}{prio}' if prio else queue
            queue_length = r.llen(queue)
            if queue_length:
                return False

    all_worker_counts = current_app.control.broadcast(
        'worker_task_count',
        reply=True,
        arguments={
            'instance': client.current_instance_name()
        }
    )

    if any([sum(r.values()) for r in all_worker_counts]):
        return False

    return True



from celery import signals, current_task

from riberry import model
from riberry.celery.client import tasks


@signals.celeryd_after_setup.connect
def setup_direct_queue(sender, instance, **kwargs):
    model.conn.raw_engine.dispose()


@signals.before_task_publish.connect
def before_task_publish(sender, headers, body, **_):
    if not current_task:
        return

    root_id = current_task.request.root_id
    args, kwargs, *_ = body
    task_id = headers['id']

    if '__ss__' in kwargs:
        tasks.create_event(
            name='stream',
            root_id=root_id,
            task_id=task_id,
            data={
                'stream': str(kwargs['__ss__']),
                'state': 'QUEUED'
            }
        )

    if '__sb__' in kwargs:
        stream, step = kwargs['__sb__']
        tasks.create_event(
            name='step',
            root_id=root_id,
            task_id=task_id,
            data={
                'stream': str(stream),
                'step': str(step),
                'state': 'QUEUED'
            }
        )


@signals.task_prerun.connect
def task_prerun(sender, kwargs, **_):
    task_id = current_task.request.id
    root_id = current_task.request.root_id
    current_task.stream = None
    current_task.step = None

    if '__ss__' in kwargs:
        stream = kwargs['__ss__']
        current_task.stream = stream
        tasks.create_event(
            name='stream',
            root_id=root_id,
            task_id=task_id,
            data={
                'stream': str(stream),
                'state': 'ACTIVE'
            }
        )

    if '__se__' in kwargs:
        stream = kwargs['__se__']
        current_task.stream = stream

    if '__sb__' in kwargs:
        stream, step = kwargs['__sb__']
        current_task.stream = stream
        current_task.step = step
        tasks.create_event(
            name='step',
            root_id=root_id,
            task_id=task_id,
            data={
                'stream': str(stream),
                'step': str(step),
                'state': 'ACTIVE'
            }
        )


@signals.task_postrun.connect
def task_postrun(sender, state, kwargs, **_):
    task_id = sender.request.id
    root_id = sender.request.root_id

    if '__ss__' in kwargs and state not in ('IGNORED', 'SUCCESS'):
        stream = kwargs['__ss__']
        tasks.create_event(
            name='stream',
            root_id=root_id,
            task_id=task_id,
            data={
                'stream': str(stream),
                'state': state or 'SS_UK'
            }
        )

    if '__se__' in kwargs:
        stream = kwargs['__se__']
        tasks.create_event(
            name='stream',
            root_id=root_id,
            task_id=task_id,
            data={
                'stream': str(stream),
                'state': state or 'SS_SE'
            }
        )

    if '__sb__' in kwargs:
        stream, step = kwargs['__sb__']
        current_task.stream = stream
        current_task.step = step
        tasks.create_event(
            name='step',
            root_id=root_id,
            task_id=task_id,
            data={
                'stream': str(stream),
                'step': str(step),
                'state': 'SUCCESS' if state == 'IGNORED' else state or 'FAILURE'
            }
        )



import datetime
import re
from typing import AnyStr, Dict, List

import jwt
import pendulum
from sqlalchemy import Column, String, ForeignKey, DateTime, desc
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.orm import relationship, validates, joinedload

from riberry import model, exc
from riberry.model import base
from riberry.config import config


class User(base.Base):
    __tablename__ = 'users'
    __reprattrs__ = ['username']

    id = base.id_builder.build()
    username = Column(String(48), nullable=False, unique=True)
    password = Column(String(512))
    auth_provider = Column(String(32), nullable=False, default=config.authentication.providers.default)
    details: 'UserDetails' = relationship('UserDetails', uselist=False, back_populates='user')

    # associations
    group_associations: List['model.group.ResourceGroupAssociation'] = model.group.ResourceGroupAssociation.make_relationship(
        resource_id=id,
        resource_type=model.misc.ResourceType.user
    )
    jobs: List['model.job.Job'] = relationship(
        'Job',
        order_by=lambda: desc(model.job.Job.created),
        back_populates='creator')
    executions: List['model.job.JobExecution'] = relationship(
        'JobExecution',
        order_by=lambda: desc(model.job.JobExecution.updated),
        back_populates='creator')
    notifications: List['model.misc.UserNotification'] = relationship(
        'UserNotification',
        order_by=lambda: desc(model.misc.UserNotification.created),
        back_populates='user')

    # proxies
    groups: List['model.group.Group'] = association_proxy('group_associations', 'group')

    @property
    def forms(self) -> List['model.interface.Form']:
        return model.interface.Form.query().filter(
            (model.group.ResourceGroupAssociation.group_id.in_(o.group_id for o in self.group_associations)) &
            (model.group.ResourceGroupAssociation.resource_type == model.misc.ResourceType.form) &
            (model.interface.Form.id == model.group.ResourceGroupAssociation.resource_id)
        ).all()

    @property
    def applications(self) -> List['model.application.Application']:
        forms = model.interface.Form.query().filter(
            model.interface.Form.id.in_(form.id for form in self.forms)
        ).options(
            joinedload(model.interface.Form.application)
        ).all()

        return [form.application for form in forms]

    @classmethod
    def authenticate(cls, username, password):
        existing_user: cls = cls.query().filter_by(username=username).first()
        if existing_user:
            provider = config.authentication[existing_user.auth_provider]
        else:
            provider = config.authentication.default_provider

        if not provider.authenticate(username=username, password=password):
            raise exc.AuthenticationError

        return cls.query().filter_by(username=username).first()

    @validates('username')
    def validate_username(self, _, username):
        if not username or len(username) < 3:
            raise ValueError(f'User.username :: usernames must be 3+ characters long. Received {repr(username)}')
        return username

    @classmethod
    def secure_password(cls, password: str, provider_name=None) -> bytes:
        provider = config.authentication[provider_name or config.authentication.providers.default]
        password = provider.secure_password(password=password)
        return password


class UserDetails(base.Base):
    __tablename__ = 'user_details'

    id = base.id_builder.build()
    user_id = Column(base.id_builder.type, ForeignKey('users.id'), nullable=False)
    user: 'User' = relationship('User', back_populates='details')

    first_name = Column(String(64))
    last_name = Column(String(64))
    display_name = Column(String(128))
    department = Column(String(128))
    email = Column(String(128))
    updated: datetime = Column(DateTime(timezone=True), default=base.utc_now)

    @property
    def full_name(self):
        if self.first_name and self.last_name:
            return f'{self.first_name} {self.last_name}'
        else:
            return self.first_name or self.last_name or self.display_name or self.user.username

    @validates('email')
    def validate_email(self, _, email):
        if email and not re.match(r'[^@]+@[^@]+\.[^@]+', email or ''):
            raise ValueError(f'UserDetails.email :: Invalid email received ({repr(email)})')
        return email


class AuthToken:

    @staticmethod
    def create(user: User, expiry_delta: datetime.timedelta=datetime.timedelta(hours=24)) -> AnyStr:
        iat: pendulum.DateTime = base.utc_now()
        exp: pendulum.DateTime = iat + expiry_delta

        return jwt.encode({
            'iat': iat.int_timestamp,
            'exp': exp.int_timestamp,
            'subject': user.username
        }, config.authentication.token.secret, algorithm='HS256')

    @staticmethod
    def verify(token: AnyStr) -> Dict:
        try:
            return jwt.decode(token, config.authentication.token.secret, algorithms=['HS256'])
        except jwt.ExpiredSignatureError:
            raise exc.SessionExpired
        except Exception:
            raise exc.AuthenticationError




import sqlalchemy
import sqlalchemy.orm
import sqlalchemy.pool

from . import misc, application, group, auth, interface, job, base


class __ModelProxy:
    raw_session = None
    raw_engine = None

    def __getattr__(self, item):
        return getattr(self.raw_session, item)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.remove()


conn: sqlalchemy.orm.session.Session = __ModelProxy()


def init(url='sqlite://', **config):
    __ModelProxy.raw_engine = sqlalchemy.create_engine(
        url,
        echo=config.get('echo', False),
        poolclass=sqlalchemy.pool.QueuePool,
        pool_use_lifo=True,
        pool_pre_ping=True,
        connect_args=config.get('connection_arguments', {})
    )
    __ModelProxy.raw_session = sqlalchemy.orm.scoped_session(sqlalchemy.orm.sessionmaker(bind=__ModelProxy.raw_engine))
    base.Base.metadata.create_all(__ModelProxy.raw_engine)



import functools
import riberry
from typing import Dict, AnyStr


class RiberryApplicationBackend:

    def __init__(self, instance):
        self.instance = instance

    def task(self, func=None, **options):
        if callable(func):
            return self.register_task(func=func, **options)
        else:
            return functools.partial(self.task, **options)

    def initialize(self):
        raise NotImplementedError

    def default_addons(self) -> Dict[AnyStr, 'riberry.app.addons.Addon']:
        raise NotImplementedError

    def register_task(self, func, **options):
        raise NotImplementedError

    def task_by_name(self, name: AnyStr):
        raise NotImplementedError

    def start_execution(self, execution_id, root_id, entry_point) -> AnyStr:
        raise NotImplementedError

    def create_receiver_task(self, external_task_id, validator):
        raise NotImplementedError

    def active_task(self):
        raise NotImplementedError

    def __getattr__(self, item):
        return getattr(self.instance, item)



from riberry import model
from datetime import timedelta
from riberry.model.group import Group


def authenticate_user(username: str, password: str) -> str:
    user = model.auth.User.authenticate(username=username, password=password)
    access_token = model.auth.AuthToken.create(user, expiry_delta=timedelta(days=5))
    return access_token.decode()


def all_groups():
    return Group.query().all()


def group_by_id(group_id):
    return Group.query().filter_by(id=group_id).first()


def users_for_group_id(group_id):
    group = group_by_id(group_id=group_id)
    return group.users


def forms_for_group_id(group_id):
    group = group_by_id(group_id=group_id)
    return group.forms


def remove_user_from_group(group_id, user_id):
    _remove_resource_from_group(group_id, user_id, model.misc.ResourceType.user)


def add_user_to_group(group_id, user_id):
    _add_resource_to_group(group_id, user_id, model.misc.ResourceType.user)


def remove_form_from_group(group_id, form_id):
    _remove_resource_from_group(group_id, form_id, model.misc.ResourceType.form)


def add_form_to_group(group_id, form_id):
    _add_resource_to_group(group_id, form_id, model.misc.ResourceType.form)


def _find_group_association(group_id, resource_id, resource_type):
    return model.group.ResourceGroupAssociation.query().filter_by(
        group_id=group_id,
        resource_id=resource_id,
        resource_type=resource_type
    ).first()


def _remove_resource_from_group(group_id, resource_id, resource_type):
    association = _find_group_association(group_id, resource_id, resource_type)
    if association:
        model.conn.delete(association)


def _add_resource_to_group(group_id, resource_id, resource_type):
    association = _find_group_association(group_id, resource_id, resource_type)
    if not association:
        association = model.group.ResourceGroupAssociation(
            group_id=group_id,
            resource_id=resource_id,
            resource_type=resource_type
        )
        model.conn.add(association)


def create_group(name):
    group = model.group.Group(name=name)
    model.conn.add(group)
    return group


import riberry
from typing import List
from .. import env


def update_all_reports(app_instance=None):
    app_instance = app_instance if app_instance else env.get_instance_model()
    reports: List[riberry.model.job.JobExecutionReport] = riberry.model.job.JobExecutionReport.query().filter(
        riberry.model.job.JobExecutionReport.marked_for_refresh == True,
    ).join(
        riberry.model.job.JobExecution
    ).join(
        riberry.model.job.Job
    ).filter_by(
        instance=app_instance
    ).all()

    cxt = env.current_context
    for report in reports:
        root_id = report.job_execution.task_id

        with cxt.scope(root_id=root_id, task_id=root_id, stream=None, step=None, category=None):
            cxt.event_registry.call(event_type=cxt.event_registry.types.on_report_refresh, report=report.name)
            if report.marked_for_refresh:
                report.marked_for_refresh = False
                riberry.model.conn.commit()



from typing import List

from celery.utils.log import logger
from sqlalchemy import desc, asc

import riberry


def _tracker_key(value):
    return f'workflow:active:{value}'


def start_tracking_execution(root_id):
    redis = riberry.celery.util.celery_redis_instance()
    instance = riberry.app.env.get_instance_name()
    key = _tracker_key(instance)
    logger.info(f'execution_tracker: Tracking workflow {root_id!r} via key {key!r}')
    redis.sadd(key, root_id)


def check_stale_execution(app_instance):
    executions: List[riberry.model.job.JobExecution] = riberry.model.job.JobExecution.query().filter(
        riberry.model.job.JobExecution.status.in_(('ACTIVE', 'READY'))
    ).join(riberry.model.job.Job).order_by(
        desc(riberry.model.job.JobExecution.priority),
        asc(riberry.model.job.JobExecution.created)
    ).filter_by(instance=app_instance).all()

    if not executions:
        return

    redis = riberry.celery.util.celery_redis_instance()
    for execution in executions:
        if not redis.sismember(_tracker_key(app_instance.internal_name), execution.task_id):
            logger.warn(
                f'execution_tracker: Identified stale workflow. '
                f'Root ID: {execution.task_id}, Execution ID:  {execution.id}'
            )

            riberry.app.actions.artifacts.create_artifact(
                name='Workflow Cancelled',
                type=riberry.model.job.ArtifactType.error,
                category='Fatal',
                filename='fatal.log',
                content=(
                    f'The current executions\'s ID ({execution.task_id}) was not found within Redis and has '
                    f'therefore been cancelled. This usually occurs when Redis is flushed while an execution is '
                    f'in the READY or ACTIVE state.'
                ),
                task_id=execution.task_id,
                root_id=execution.task_id,
            )

            riberry.app.actions.executions.execution_complete(
                task_id=execution.task_id,
                root_id=execution.task_id,
                status='FAILURE',
                stream=None
            )



from typing import List

from celery import current_app
from celery.utils.log import logger

from riberry import model
from riberry.celery import client


class DynamicParameter:

    def __init__(self, parameter):
        self.parameter = parameter

    def on_received(self, instance: model.application.ApplicationInstance, value: str):
        raise NotImplementedError


def make_dynamic_parameters_task(handlers: List[DynamicParameter]):

    def dynamic_parameters_task():
        if not handlers:
            return

        with model.conn:
            instance: model.application.ApplicationInstance = model.application.ApplicationInstance.query().filter_by(
                internal_name=client.current_instance_name(raise_on_none=True)
            ).one()
            active_schedules = instance.active_schedules

        handler_mapping = {h.parameter: h for h in handlers}
        for parameter, schedule in sorted(active_schedules.items(), key=lambda item: -item[1].priority if item[1] else 0):
            if parameter in handler_mapping:
                value = schedule.value if schedule else None
                logger.info(f'dynamic-parameters: updating dynamic parameter {parameter!r} with value {value!r}')
                try:
                    handler_mapping[parameter].on_received(instance=instance, value=value)
                except:
                    logger.exception(f'dynamic-parameters: {parameter!r} failed')

    dynamic_parameters_task.__name__ = 'dynamic'
    return dynamic_parameters_task


class DynamicParameters:
    all_apps = {}

    def __init__(self, riberry_workflow, handlers, beat_queue):
        self.workflow = riberry_workflow
        self.all_apps[self.workflow.app.main] = self
        self.workflow.app.task(name='dynamic-parameters', rib_task=False)(make_dynamic_parameters_task(handlers=handlers))
        self._configure_beat_queues(self.workflow.app, beat_queue)

    @staticmethod
    def _configure_beat_queues(app, beat_queue):
        schedule = {
            'dynamic-parameters': {
                'task': 'dynamic-parameters',
                'schedule': 2,
                'options': {'queue': beat_queue}
            },
        }

        if not app.conf.beat_schedule:
            app.conf.beat_schedule = {}
        app.conf.beat_schedule.update(schedule)

    @classmethod
    def instance(cls) -> 'DynamicParameters':
        return cls.all_apps.get(current_app.main)



import os
from celery.worker import control


@control.control_command()
def toggle_external_task_queue(state, operation):
    if os.environ.get('RIBERRY_EXTERNAL_TASK'):

        queues = {q.name for q in state.consumer.task_consumer.queues}
        if operation == 'add' and 'rib.manual' not in queues:
            state.consumer.add_task_queue('rib.manual')
            return True

        if operation == 'cancel' and 'rib.manual' in queues:
            state.consumer.cancel_task_queue('rib.manual')
            return True

    return None



import abc
import functools
import inspect
from contextlib import contextmanager
from threading import local
from typing import Set, Union, Optional, Type, Callable

from riberry.exc import AuthorizationError


class NotApplicable(Exception):
    pass


class AuthorizationEngine:

    def __init__(self, name, *policy):
        self.name = name
        self.policies: Set[Union[PolicySet, Policy]] = set(policy)

    def authorize(self, context):
        results = []
        for policy_sey in self.policies:
            try:
                results.append(policy_sey.authorize(context=context))
            except NotApplicable:
                pass

        return False if False in results else True


class PolicyContext:

    _local = local()
    _no_default = object()

    def __getitem__(self, item):
        return getattr(self._local, item, None)

    def __setitem__(self, item, value):
        setattr(self._local, item, value)

    @property
    def enabled(self):
        return self['enabled']

    @enabled.setter
    def enabled(self, value):
        self['enabled'] = value

    @property
    def subject(self):
        return self['subject']

    @subject.setter
    def subject(self, value):
        self['subject'] = value

    @property
    def environment(self):
        return self['environment']

    @environment.setter
    def environment(self, value):
        self['environment'] = value

    @property
    def engine(self) -> AuthorizationEngine:
        return self['policy_engine']

    @engine.setter
    def engine(self, value):
        self['policy_engine'] = value

    @property
    def on_deny(self) -> Optional[Exception]:
        return self['on_deny']

    @on_deny.setter
    def on_deny(self, value):
        self['on_deny'] = value

    @contextmanager
    def scope(self, subject, environment, policy_engine, on_deny: Optional[Union[Type, Callable]] = AuthorizationError):
        self.configure(subject=subject, environment=environment, policy_engine=policy_engine, on_deny=on_deny)
        yield
        self.reset()

    @contextmanager
    def disabled_scope(self):
        try:
            self.enabled = False
            yield
        finally:
            self.enabled = True

    def configure(self, subject, environment, policy_engine, on_deny: Optional[Union[Type, Callable]] = AuthorizationError):
        self.enabled = True
        self.subject = subject
        self.environment = environment
        self.engine = policy_engine
        self.on_deny = on_deny

    def reset(self):
        self.enabled = True
        self.subject = None
        self.environment = None
        self.engine = None
        self.on_deny = None

    @classmethod
    def current(cls):
        return cls()

    def authorize(self, resource, action, on_deny: Optional[Union[Type, Callable]] = _no_default):

        if not self.enabled:
            return True

        attr_context = AttributeContext(
            subject=self.subject,
            environment=self.environment,
            action=action,
            resource=resource
        )

        result = self.engine.authorize(attr_context)
        if result is False:
            on_deny = self.on_deny if on_deny is self._no_default else on_deny
            if inspect.isclass(on_deny) and issubclass(on_deny, Exception):
                raise on_deny
            elif callable(on_deny):
                on_deny(attr_context)
        return result

    def filter(self, resources, action):
        return [resource for resource in resources if self.authorize(resource, action, on_deny=None)]

    def post_filter(self, action):
        def outer(func):
            @functools.wraps(func)
            def inner(*args, **kwargs):
                result = func(*args, **kwargs)
                return self.filter(resources=result, action=action)
            return inner
        return outer

    def post_authorize(self, action, on_deny: Optional[Union[Type, Callable]] = _no_default):
        def outer(func):
            @functools.wraps(func)
            def inner(*args, **kwargs):
                result = func(*args, **kwargs)
                self.authorize(resource=result, action=action, on_deny=on_deny)
                return result
            return inner
        return outer


class AttributeContext:

    def __init__(self, subject, resource, action, environment):
        self.subject = subject
        self.resource = resource
        self.action = action
        self.environment = environment


class AuthorizationElement(metaclass=abc.ABCMeta):

    def target_clause(self, context: AttributeContext) -> bool:
        raise NotImplementedError

    def condition(self, context: AttributeContext) -> bool:
        raise NotImplementedError

    def apply(self, context: AttributeContext):
        if not self.target_clause(context=context):
            raise NotApplicable
        return self.condition(context=context)

    def authorize(self, context):
        raise NotImplementedError


class PolicyCollection(AuthorizationElement, metaclass=abc.ABCMeta):

    def __init__(self, *collection):
        self.collection: Set[AuthorizationElement] = set(collection)

    def authorize(self, context):
        if not self.apply(context=context):
            return False

        results = set()
        for policy in self.collection:
            try:
                results.add(policy.authorize(context=context))
            except NotApplicable:
                pass

        if any(results):
            return True
        return False if False in results else None


class PolicySet(PolicyCollection, metaclass=abc.ABCMeta):
    pass


class Policy(PolicyCollection, metaclass=abc.ABCMeta):
    pass


class Rule(AuthorizationElement, metaclass=abc.ABCMeta):

    def on_permit(self, context):
        pass

    def on_deny(self, context):
        pass

    def authorize(self, context):
        result = self.apply(context=context)
        self.on_permit(context=context) if result else self.on_deny(context=context)
        return result



import enum
import functools
import json
import mimetypes
from datetime import datetime
from typing import List, Optional

import pendulum
from croniter import croniter
from sqlalchemy import Column, String, ForeignKey, DateTime, Boolean, Integer, Binary, Index, Enum, desc, Float, asc, \
    UniqueConstraint, select, func, sql
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import relationship, deferred, validates, foreign, remote

from riberry import model
from riberry.model import base


class ArtifactType(enum.Enum):
    output = 'output'
    error = 'error'
    report = 'report'

    def __repr__(self):
        return repr(self.value)


class Job(base.Base):
    """
    A Job is an object which represents a set of inputs provided to a form. A Job can have one or more JobExecutions
    which represent the execution of the linked ApplicationInterface on the linked ApplicationInstance for the given
    input stored against this Job.

    Jobs are immutable. If we require a different set of values, we'll need to create a new Job.
    """

    __tablename__ = 'job'
    __reprattrs__ = ['name']
    __table_args__ = (
        Index('j__idx_form_id', 'form_id'),
        Index('j__idx_creator_id', 'creator_id'),
    )

    # columns
    id = base.id_builder.build()
    form_id = Column(base.id_builder.type, ForeignKey('form.id'), nullable=False)
    creator_id = Column(base.id_builder.type, ForeignKey('users.id'), nullable=False)
    name: str = Column(String(64), nullable=False, unique=True, comment='The unique name of our job.')
    created: datetime = Column(DateTime(timezone=True), default=base.utc_now, nullable=False)

    # associations
    creator: 'model.auth.User' = relationship('User')
    form: 'model.interface.Form' = relationship('Form', back_populates='jobs')
    executions: List['JobExecution'] = relationship(
        'JobExecution',
        cascade='save-update, merge, delete, delete-orphan',
        order_by=lambda: desc(JobExecution.updated),
        back_populates='job')
    schedules: List['JobSchedule'] = relationship('JobSchedule', cascade='save-update, merge, delete, delete-orphan',
                                                  back_populates='job')
    values: List['model.interface.InputValueInstance'] = relationship('InputValueInstance',
                                                                      cascade='save-update, merge, delete, delete-orphan',
                                                                      back_populates='job')
    files: List['model.interface.InputFileInstance'] = relationship('InputFileInstance',
                                                                    cascade='save-update, merge, delete, delete-orphan',
                                                                    back_populates='job')

    # proxies
    instance: 'model.application.ApplicationInstance' = association_proxy('form', 'instance')

    def execute(self, creator_id):
        model.conn.add(instance=JobExecution(job=self, creator_id=creator_id))


class JobSchedule(base.Base):
    """
    A JobSchedule object defines a schedule which will trigger an execution for the linked Job.

    Note that a JobExecution will only be created for an ApplicationInstance which has a status of "online".
    Applications which are offline or inactive due to an ApplicationInstanceSchedule will not have JobExecutions
    created.
    """

    __tablename__ = 'sched_job'

    # columns
    id = base.id_builder.build()
    job_id = Column(base.id_builder.type, ForeignKey('job.id'), nullable=False)
    creator_id = Column(base.id_builder.type, ForeignKey('users.id'), nullable=False)
    enabled: bool = Column(Boolean, default=True, nullable=False, comment='Whether or not this schedule is active.')
    cron: str = Column(String(24), nullable=False, comment='The cron expression which defines our schedule.')
    created: datetime = Column(DateTime(timezone=True), default=base.utc_now, nullable=False,
                               comment='The time our schedule was created.')
    last_run: datetime = Column(DateTime(timezone=True), default=None,
                                comment='The last time a job execution was created from our schedule.')
    limit: int = Column(Integer, default=0, comment='The amount of valid runs for this schedule.')
    total_runs: int = Column(Integer, default=0, comment='The total amount of runs for this schedule.')

    # associations
    job: 'Job' = relationship('Job', back_populates='schedules')
    creator: 'model.auth.User' = relationship('User')

    def run(self):
        if not self.enabled or self.job.instance.status != 'online':
            return

        ready_run = None
        for cron_time in croniter(self.cron, start_time=self.last_run or self.created, ret_type=pendulum.DateTime):
            cron_time = pendulum.instance(cron_time)
            if cron_time > base.utc_now():
                break
            ready_run = cron_time

        if ready_run:
            self.last_run = ready_run
            self.total_runs += 1
            if self.limit and self.total_runs >= self.limit:
                self.enabled = False

            self.job.execute(creator_id=self.creator_id)

    @property
    def next_run(self):
        if not self.enabled:
            return

        instance = croniter(self.cron, start_time=self.last_run or self.created, ret_type=pendulum.DateTime)
        return pendulum.instance(next(instance))


def memoize(func):
    result = []
    @functools.wraps(func)
    def inner(*args, **kwargs):
        if result:
            return result[0]
        result.append(func(*args, **kwargs))
        return result[0]
    return inner


@memoize
def _job_execution_select_latest_progress():
    return select([
        func.max(JobExecutionProgress.id).label('id'),
        JobExecutionProgress.job_execution_id
    ]).group_by(
        JobExecutionProgress.job_execution_id
    ).alias()


class JobExecution(base.Base):
    """A JobExecution represent a single execution of our Job."""
    __tablename__ = 'job_execution'
    __reprattrs__ = ['job_id', 'task_id', 'status']
    __table_args__ = (
        Index('j_e__idx_job_id', 'job_id'),
        Index('j_e__idx_creator_id', 'creator_id'),
    )

    # columns
    id = base.id_builder.build()
    job_id = Column(base.id_builder.type, ForeignKey('job.id'), nullable=False)
    creator_id = Column(base.id_builder.type, ForeignKey('users.id'), nullable=False)
    task_id: str = Column(String(36), unique=True,
                          comment='The internal identifier of our job execution. This is usually the Celery root ID.')
    status: str = Column(String(24), default='RECEIVED', comment='The current status of our job execution.')
    created: datetime = Column(DateTime(timezone=True), default=base.utc_now, nullable=False)
    started: datetime = Column(DateTime(timezone=True))
    completed: datetime = Column(DateTime(timezone=True))
    updated: datetime = Column(DateTime(timezone=True), default=base.utc_now, nullable=False)
    priority = Column(Integer, default=64, nullable=False,
                      comment='The priority of this execution. This only applies to tasks in the RECEIVED state.')
    parent_execution_id = Column(base.id_builder.type, ForeignKey('job_execution.id'),
                                 comment='The id of the execution which triggered this execution.')

    # associations
    creator: 'model.auth.User' = relationship('User')
    job: 'Job' = relationship('Job', back_populates='executions')
    streams: List['JobExecutionStream'] = relationship(
        'JobExecutionStream',
        cascade='save-update, merge, delete, delete-orphan',
        order_by=lambda: asc(JobExecutionStream.id),
        back_populates='job_execution'
    )
    artifacts: List['JobExecutionArtifact'] = relationship(
        'JobExecutionArtifact',
        cascade='save-update, merge, delete, delete-orphan',
        order_by=lambda: asc(JobExecutionArtifact.id),
        back_populates='job_execution')
    external_tasks: List['JobExecutionExternalTask'] = relationship(
        'JobExecutionExternalTask',
        cascade='save-update, merge, delete, delete-orphan',
        order_by=lambda: asc(JobExecutionExternalTask.id),
        back_populates='job_execution'
    )
    reports: List['JobExecutionReport'] = relationship(
        'JobExecutionReport',
        cascade='save-update, merge, delete, delete-orphan',
        order_by=lambda: asc(JobExecutionReport.id),
        back_populates='job_execution'
    )
    progress: List['JobExecutionProgress'] = relationship(
        'JobExecutionProgress',
        cascade='save-update, merge, delete, delete-orphan',
        order_by=lambda: JobExecutionProgress.id.asc(),
        back_populates='job_execution'
    )
    data: List['model.misc.ResourceData'] = model.misc.ResourceData.make_relationship(
        resource_id=id,
        resource_type=model.misc.ResourceType.job_execution,
    )

    latest_progress: 'JobExecutionProgress' = relationship(
        'JobExecutionProgress',
        secondary=lambda: _job_execution_select_latest_progress(),
        primaryjoin=lambda: JobExecution.id == _job_execution_select_latest_progress().c.job_execution_id,
        secondaryjoin=lambda: JobExecutionProgress.id == _job_execution_select_latest_progress().c.id,
        viewonly=True,
        uselist=False,
    )

    parent_execution: 'JobExecution' = relationship('JobExecution', back_populates='child_executions', remote_side=[id])
    child_executions: List['JobExecution'] = relationship('JobExecution', back_populates='parent_execution')

    # validations
    @validates('priority')
    def validate_priority(self, _, priority):
        assert isinstance(priority, int) and 255 >= priority >= 1, (
            f'ApplicationInstanceSchedule.priority must be an integer between 1 and 255 (received {priority})')
        return priority

    @property
    def stream_status_summary(self):
        summary = model.conn.query(
            model.job.JobExecutionStream.status,
            func.count(model.job.JobExecutionStream.status)
        ).filter_by(
            job_execution=self,
        ).group_by(
            model.job.JobExecutionStream.status,
        ).all()

        return dict(summary)


class JobExecutionProgress(base.Base):
    __tablename__ = 'job_progress'
    __reprattrs__ = ['message']

    # columns
    id = base.id_builder.build()
    job_execution_id = Column(base.id_builder.type, ForeignKey('job_execution.id'), nullable=False)
    created: datetime = Column(DateTime(timezone=True), default=base.utc_now, nullable=False)
    message: str = Column(String(256), default=None, comment='Message describing the progress of the job execution.')
    progress_percentage = Column(Integer, default=None, nullable=True, comment='The progress of the job execution.')

    # associations
    job_execution: 'JobExecution' = relationship('JobExecution', back_populates='progress')

    # validations
    @validates('progress_percentage')
    def validate_priority(self, _, progress_percentage):
        if progress_percentage is not None:
            progress_percentage = min(max(progress_percentage, 0), 100)
        return progress_percentage


class JobExecutionStream(base.Base):
    __tablename__ = 'job_stream'
    __reprattrs__ = ['name', 'task_id', 'status']
    __table_args__ = (
        Index('j_s__idx_job_execution_id', 'job_execution_id'),
    )

    # columns
    id = base.id_builder.build()
    job_execution_id = Column(base.id_builder.type, ForeignKey('job_execution.id'), nullable=False)
    task_id: str = Column(String(36), unique=True)
    parent_stream_id = Column(base.id_builder.type, ForeignKey('job_stream.id'))
    name: str = Column(String(64), nullable=False)
    category: str = Column(String(64), nullable=False, default='Overall')
    status: str = Column(String(24), default='QUEUED')
    created: datetime = Column(DateTime(timezone=True), default=base.utc_now, nullable=False)
    started: datetime = Column(DateTime(timezone=True))
    completed: datetime = Column(DateTime(timezone=True))
    updated: datetime = Column(DateTime(timezone=True), default=base.utc_now, nullable=False)

    # associations
    job_execution: 'JobExecution' = relationship('JobExecution', back_populates='streams')
    steps: List['JobExecutionStreamStep'] = relationship(
        'JobExecutionStreamStep',
        cascade='save-update, merge, delete, delete-orphan',
        order_by=lambda: asc(JobExecutionStreamStep.id),
        back_populates='stream',
    )
    artifacts: List['JobExecutionArtifact'] = relationship('JobExecutionArtifact', back_populates='stream')
    external_tasks: List['JobExecutionExternalTask'] = relationship(
        'JobExecutionExternalTask',
        cascade='save-update, merge, delete, delete-orphan',
        order_by=lambda: asc(JobExecutionExternalTask.id),
        back_populates='stream',
    )
    parent_stream: 'JobExecutionStream' = relationship('JobExecutionStream', back_populates='child_streams',
                                                       remote_side=[id])
    child_streams: List['JobExecutionStream'] = relationship('JobExecutionStream', back_populates='parent_stream')


class JobExecutionStreamStep(base.Base):
    __tablename__ = 'job_stream_step'
    __reprattrs__ = ['name', 'task_id', 'status']
    __table_args__ = (
        Index('j_s_s__idx_stream_id', 'stream_id'),
    )

    # columns
    id = base.id_builder.build()
    stream_id = Column(base.id_builder.type, ForeignKey('job_stream.id'), nullable=False)
    task_id: str = Column(String(36), unique=True)
    name: str = Column(String(64))
    status: str = Column(String(24), default='RECEIVED')
    created: datetime = Column(DateTime(timezone=True), default=base.utc_now, nullable=False)
    started: datetime = Column(DateTime(timezone=True))
    completed: datetime = Column(DateTime(timezone=True))
    updated: datetime = Column(DateTime(timezone=True), default=base.utc_now, nullable=False)

    # associations
    stream: 'JobExecutionStream' = relationship('JobExecutionStream', back_populates='steps')


class JobExecutionArtifact(base.Base):
    __tablename__ = 'job_artifact'
    __reprattrs__ = ['name', 'filename']
    __table_args__ = (
        Index('j_a__idx_job_execution_id', 'job_execution_id'),
        Index('j_a__idx_stream_id', 'stream_id'),
    )

    # columns
    id = base.id_builder.build()
    job_execution_id = Column(base.id_builder.type, ForeignKey('job_execution.id'), nullable=False)
    stream_id = Column(base.id_builder.type, ForeignKey('job_stream.id'))
    name: str = Column(String(128), nullable=False)
    type: str = Column(Enum(ArtifactType), nullable=False)
    category: str = Column(String(128), nullable=False, default='Default')
    filename: str = Column(String(512), nullable=False)
    created: datetime = Column(DateTime(timezone=True), default=base.utc_now, nullable=False)
    size: int = Column(Integer, nullable=False)

    # associations
    job_execution: 'JobExecution' = relationship('JobExecution', back_populates='artifacts')
    stream: 'JobExecutionStream' = relationship('JobExecutionStream', back_populates='artifacts')
    binary: 'JobExecutionArtifactBinary' = relationship(
        'JobExecutionArtifactBinary', cascade='save-update, merge, delete, delete-orphan', back_populates='artifact',
        uselist=False)
    data: List['JobExecutionArtifactData'] = relationship(
        'JobExecutionArtifactData', cascade='save-update, merge, delete, delete-orphan', back_populates='artifact')

    @property
    def content_type(self):
        if self.filename.endswith('.log'):  # quick fix for missing .log type on unix systems
            return 'text/plain'

        return mimetypes.guess_type(self.filename)[0]

    @property
    def content_encoding(self):
        return mimetypes.guess_type(self.filename)[1]


class JobExecutionArtifactData(base.Base):
    __tablename__ = 'job_artifact_data'

    # columns
    id = base.id_builder.build()
    title: str = Column(String(64), nullable=False)
    description: str = Column(String(512), nullable=False)
    artifact_id = Column(base.id_builder.type, ForeignKey('job_artifact.id'), nullable=False)

    # associations
    artifact: 'JobExecutionArtifact' = relationship('JobExecutionArtifact', back_populates='data')


class JobExecutionArtifactBinary(base.Base):
    __tablename__ = 'job_artifact_binary'

    # columns
    id = base.id_builder.build()
    binary: bytes = deferred(Column(Binary, nullable=True))
    artifact_id = Column(base.id_builder.type, ForeignKey('job_artifact.id'), nullable=False)

    # associations
    artifact: 'JobExecutionArtifact' = relationship('JobExecutionArtifact', back_populates='binary')


class JobExecutionExternalTask(base.Base):
    __tablename__ = 'job_external_task'
    __reprattrs__ = ['name', 'type']
    __table_args__ = (
        Index('j_e__idx_job_execution_id', 'job_execution_id'),
        Index('j_e__idx_stream_id', 'stream_id'),
    )

    # columns
    id = base.id_builder.build()
    job_execution_id = Column(base.id_builder.type, ForeignKey('job_execution.id'), nullable=False)
    stream_id = Column(base.id_builder.type, ForeignKey('job_stream.id'))
    user_id = Column(base.id_builder.type, ForeignKey('users.id'), nullable=True)
    group_id = Column(base.id_builder.type, ForeignKey('groups.id'), nullable=True)
    task_id: str = Column(String(64), unique=True)
    name: str = Column(String(128), nullable=False)
    type: str = Column(String(24), nullable=False)
    status: str = Column(String(24), default='WAITING', comment='The current status of the manual task.')
    input_data: Optional[bytes] = deferred(Column(Binary, nullable=True))
    output_data: Optional[bytes] = deferred(Column(Binary, nullable=True))

    # associations
    job_execution: 'JobExecution' = relationship('JobExecution', back_populates='external_tasks')
    stream: 'JobExecutionStream' = relationship('JobExecutionStream', back_populates='external_tasks')


class JobExecutionReport(base.Base):
    __tablename__ = 'job_report'
    __reprattrs__ = ['internal_name']

    # columns
    id = base.id_builder.build()
    job_execution_id = Column(base.id_builder.type, ForeignKey('job_execution.id'), nullable=False)
    name: str = Column(String(128), nullable=False)
    renderer: str = Column(String(24), nullable=False, default='unspecified')
    title: str = Column(String(128), nullable=True)
    category: str = Column(String(128), nullable=True)
    key: str = Column(String(128), nullable=True)
    raw_input_data: Optional[bytes] = deferred(Column('input_data', Binary, nullable=True))
    report: Optional[bytes] = deferred(Column(Binary, nullable=True))
    marked_for_refresh: bool = Column(Boolean, nullable=False, default=False)

    # associations
    job_execution: 'JobExecution' = relationship('JobExecution', back_populates='reports')

    @hybrid_property
    def input_data(self):
        return json.loads(self.raw_input_data.decode()) if self.raw_input_data else None

    @input_data.setter
    def input_data(self, value):
        self.raw_input_data = json.dumps(value).encode()




import abc


class AuthenticationProvider(metaclass=abc.ABCMeta):

    def __init__(self, config_dict):
        self.raw_config = config_dict

    @classmethod
    def name(cls) -> str:
        raise NotImplementedError

    def authenticate(self, username: str, password: str) -> bool:
        raise NotImplementedError

    def secure_password(self, password: bytes) -> bytes:
        raise NotImplementedError

    def on_enabled(self):
        pass



from contextlib import contextmanager

from celery import current_app
from celery.utils.log import logger

from riberry.celery import client
from riberry.celery.client.dynamic import DynamicParameter
from riberry.celery.client.dynamic.util import PriorityQueue
from riberry.celery.client.scale import redis_queues_empty_workers_idle, ConcurrencyScale
from riberry.celery.util import celery_redis_instance


class DynamicQueues(DynamicParameter):

    def __init__(self, parameter='active'):
        super(DynamicQueues, self).__init__(parameter=parameter)

    def on_received(self, instance, value):
        scale = ConcurrencyScale.instance()
        value = str(value).upper()

        if value == 'N':
            logger.info('DynamicQueues: app down')
            current_app.control.broadcast('scale_down', arguments={'instance': client.current_instance_name()})
        elif scale and redis_queues_empty_workers_idle(scale.target_queues):
            logger.info('DynamicQueues: empty queues')
            current_app.control.broadcast('scale_down', arguments={'instance': client.current_instance_name()})
        elif value == 'Y':
            logger.info('DynamicQueues: scaling up')
            current_app.control.broadcast('scale_up', arguments={'instance': client.current_instance_name()})


class DynamicConcurrency(DynamicParameter):

    def __init__(self, parameter='concurrency'):
        super(DynamicConcurrency, self).__init__(parameter=parameter)

    def on_received(self, instance, value):
        current_app.control.broadcast('scale_to', arguments={
            'concurrency': int(value),
            'instance': client.current_instance_name()
        })


class DynamicPriorityParameter(client.dynamic.DynamicParameter):

    def __init__(
            self, parameter='hosts', key=None, sep='|', queue_cls=PriorityQueue,
            blocking: bool=True, block_retry: int=0.5, r=None):
        super(DynamicPriorityParameter, self).__init__(parameter=parameter)
        self.r = r or celery_redis_instance()
        self.sep = sep
        self.queue = queue_cls(
            r=self.r,
            key=key or client.current_instance_name(raise_on_none=True),
            blocking=blocking,
            block_retry=block_retry,
        )

    @property
    def last_value_key(self):
        return self.queue.make_key(self.queue.key, 'raw')

    @property
    def last_value(self):
        last_value = self.r.get(self.last_value_key)
        return last_value.decode() if isinstance(last_value, bytes) else None

    @last_value.setter
    def last_value(self, value):
        self.r.set(self.last_value_key, value=value)

    def on_received(self, instance, value):
        if self.last_value is not None and value == self.last_value:
            return

        self.last_value = value
        values = [
            part.split(self.sep) if self.sep in part else (part, 0)
            for part in (value or '').split(' ')
        ]

        member_scores = {k: int(v) for k, v in values}
        self.queue.update(member_scores)
        logger.info(f'DynamicPriorityParameter: ({self.queue.free_key}) updated {self.parameter} queue with {value!r}')

    @contextmanager
    def borrow(self):
        member, score, version = self.queue.pop()
        try:
            yield member, score, version
        finally:
            self.queue.put(member=member, version=version)



from .base import AddonStartStopStep
import riberry


class BackgroundTasks(riberry.app.addons.Addon):

    def register(self, riberry_app: 'riberry.app.base.RiberryApplication'):
        class ConcreteBackgroundTasksStep(BackgroundTasksStep):
            rib = riberry_app

        riberry_app.backend.steps['worker'].add(ConcreteBackgroundTasksStep)


class BackgroundTasksStep(AddonStartStopStep):
    requires = {'celery.worker.components:Timer'}

    def __init__(self, worker, **_):
        super().__init__(worker=worker, interval=0.6)
        self.lock = riberry.app.util.redis_lock.RedisLock(name='step:background', on_acquired=self.on_lock_acquired, interval=500)

    @staticmethod
    def on_lock_acquired():
        try:
            riberry.app.tasks.echo()
            riberry.app.tasks.poll()
            riberry.app.tasks.refresh()
        finally:
            riberry.model.conn.remove()
            riberry.model.conn.raw_engine.dispose()

    def should_run(self) -> bool:
        return True

    def run(self):
        redis_instance = riberry.celery.util.celery_redis_instance()
        self.lock.run(redis_instance=redis_instance)



import json

import pendulum

import riberry


def create_event(name, root_id, task_id, data=None, binary=None):
    if not root_id:
        return

    if isinstance(binary, str):
        binary = binary.encode()

    evt = riberry.model.misc.Event(
        name=name,
        time=pendulum.DateTime.utcnow().timestamp(),
        task_id=task_id,
        root_id=root_id,
        data=json.dumps(data),
        binary=binary,
    )

    riberry.model.conn.add(evt)
    riberry.model.conn.commit()



import enum
from datetime import datetime
from typing import List

import pendulum
from sqlalchemy import Column, String, Boolean, ForeignKey, DateTime, Integer, desc, Enum
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.orm import relationship, validates

from riberry import model
from riberry.model import base


class Application(base.Base):
    """Contains basic metadata related to an application."""

    __tablename__ = 'application'
    __reprattrs__ = ['name', 'enabled']

    # columns
    id = base.id_builder.build()
    document_id = Column(base.id_builder.type, ForeignKey(column='document.id'))
    name: str = Column(String(64), nullable=False, unique=True, comment='The human-readable name of the application.')
    internal_name: str = Column(String(256), nullable=False, unique=True, comment='The internal name or secondary identifier of the application.')
    description: str = Column(String(256), comment='A brief description of the application\'s purpose.')
    type: str = Column(String(64), nullable=False, comment='The type of application.')
    enabled: bool = Column(Boolean, default=True, comment='Whether or not this application and its instances are enabled (TODO).')

    # associations
    instances: List['ApplicationInstance'] = relationship(
        'ApplicationInstance', cascade='save-update, merge, delete, delete-orphan', back_populates='application')
    forms: List['model.interface.Form'] = relationship(
        'Form', cascade='save-update, merge, delete, delete-orphan', back_populates='application')
    document: 'model.misc.Document' = relationship('Document', cascade='save-update, merge, delete, delete-orphan', single_parent=True)


class ApplicationInstance(base.Base):
    """
    An ApplicationInstance represents a running instance of an Application. Multiple ApplicationInstances are useful
    when we want to separate out different types of executions for the same Application.

    An example of this is when we have an application which can support both long-running and short-running executions.
    We can spin up a separate instance with separate scheduling to ensure that the long-running jobs don't consume block
    the short-running jobs.
    """

    __tablename__ = 'app_instance'
    __reprattrs__ = ['name', 'internal_name']

    # columns
    id = base.id_builder.build()
    application_id = Column(base.id_builder.type, ForeignKey(column='application.id'), nullable=False)
    name: str = Column(String(64), nullable=False, unique=True, comment='The human-readable name of the application.')
    internal_name: str = Column(String(256), nullable=False, unique=True, comment='The internal name or secondary identifier of the application instance.')

    # associations
    application: 'Application' = relationship('Application', back_populates='instances')
    heartbeat: 'Heartbeat' = relationship('Heartbeat', cascade='save-update, merge, delete, delete-orphan', uselist=False, back_populates='instance')
    schedules: List['ApplicationInstanceSchedule'] = relationship(
        'ApplicationInstanceSchedule',
        cascade='save-update, delete, delete-orphan',
        back_populates='instance',
        order_by=lambda: (
            ApplicationInstanceSchedule.parameter,
            desc(ApplicationInstanceSchedule.priority),
            ApplicationInstanceSchedule.start_time,
        )
    )
    forms: List['model.interface.Form'] = relationship('Form', cascade='save-update, merge, delete, delete-orphan', back_populates='instance')

    @property
    def status(self):
        if not self.heartbeat:
            return 'created'

        diff = base.utc_now() - pendulum.instance(self.heartbeat.updated)
        if diff.seconds >= 10:
            return 'offline'

        if self.active_schedule_value('active', default='Y') == 'N':
            return 'inactive'

        return 'online'

    def active_schedule_value(self, name, default=None, current_time=None):
        schedule = self.active_schedule(name=name, current_time=current_time)
        return schedule.value if schedule else default

    def active_schedule(self, name, current_time=None) -> 'ApplicationInstanceSchedule':
        schedules = sorted(
            (s for s in self.schedules if s.parameter == name and s.active(current_time=current_time)),
            key=lambda s: (-s.priority, s.start_time)
        )

        for schedule in schedules:
            return schedule

    @property
    def active_schedule_values(self):
        return {name: schedule.value if schedule else None for name, schedule in self.active_schedules.items()}

    @property
    def active_schedules(self):
        parameters = {s.parameter for s in self.schedules}
        return {param: self.active_schedule(name=param) for param in parameters}


class Heartbeat(base.Base):
    """
    The Heartbeat object is used by running ApplicationInstances to report back that they're alive. If an
    ApplicationInstance's last heartbeat was over 10 seconds ago, we assume that it is offline.
    """

    __tablename__ = 'heartbeat_app_instance'
    __reprattrs__ = ['instance_id', 'updated']

    # columns
    id = base.id_builder.build()
    instance_id = Column(base.id_builder.type, ForeignKey('app_instance.id'), nullable=False)
    created: datetime = Column(DateTime(timezone=True), default=base.utc_now, comment='The first heartbeat we received for the instance.')
    updated: datetime = Column(DateTime(timezone=True), default=base.utc_now, comment='The last heartbeat we received for the instance.')

    # associations
    instance: 'ApplicationInstance' = relationship('ApplicationInstance', back_populates='heartbeat')


class ApplicationInstanceSchedule(base.Base):
    """
    The ApplicationInstanceSchedule is a time-based schedule for dynamic parameters. It allows us to define parameters
    which change depending on the time of the day.

    There are currently two core parameters: `active` and `concurrency`. These allow us to automatically
    activate/de-activate and scale up/down our applications during different times of the day.

    Any custom parameter can be defined and consumed by our Celery application as long as we have a custom parameter
    handler in place which can interpret the parameter's values.
    """

    __tablename__ = 'sched_app_instance'

    # columns
    id = base.id_builder.build()
    instance_id = Column(base.id_builder.type, ForeignKey('app_instance.id'), nullable=False)
    days = Column(String(27), nullable=False, default='*', comment='Comma-separated list of specific days ("MON,WED"), or "*" for every day.')
    start_time = Column(String(8), nullable=False, default='00:00:00', comment='The time when this schedule activates.')
    end_time = Column(String(8), nullable=False, default='23:59:59', comment='The time when this schedule de-activates.')
    timezone = Column(String(128), nullable=False, default='UTC', comment='The timezone of for the given start and end times.')
    parameter = Column(String(32), nullable=False, comment='The parameter which this schedule applies to.')
    value = Column(String(512), comment='The value of the given parameter.')
    priority = Column(Integer, default=64, nullable=False, comment='Priority of the schedule, where higher values mean higher priority.')

    # associations
    instance: 'ApplicationInstance' = relationship('ApplicationInstance', back_populates='schedules')

    # validations
    @validates('priority')
    def validate_priority(self, _, priority):
        assert isinstance(priority, int) and 255 >= priority >= 1, (
            f'ApplicationInstanceSchedule.priority must be an integer between 1 and 255 (received {priority})')
        return priority

    @validates('start_time')
    def validate_start_time(self, _, start_time):
        return self.cleanse_time(time_=start_time)

    @validates('end_time')
    def validate_end_time(self, _, end_time):
        return self.cleanse_time(time_=end_time)

    @staticmethod
    def cleanse_time(time_):
        if isinstance(time_, int):
            now = pendulum.now()
            date = pendulum.DateTime(now.year, now.month, now.day) + pendulum.duration(seconds=time_)
            time_ = date.strftime('%H:%M:%S')

        if isinstance(time_, str):
            pendulum.DateTime.strptime(time_, '%H:%M:%S')
        return time_

    @validates('timezone')
    def validate_timezone(self, _, timezone):
        pendulum.timezone(timezone)
        return timezone

    @validates('days')
    def validate_days(self, _, days):
        if days in (None, '*'):
            return '*'

        days = [o.strip() for o in str(days).lower().split(',')]
        invalid_days = set(days) - {'mon', 'tue', 'wed', 'thu', 'fri', 'sat', 'sun'}
        if invalid_days:
            raise ValueError(f'Invalid days received: {", ".join(invalid_days)}')
        return ','.join(days)

    def active(self, current_time=None):
        now: pendulum.DateTime = pendulum.instance(current_time) \
            if current_time else pendulum.DateTime.now(tz=self.timezone)
        now = now.replace(microsecond=0)

        if self.days != '*':
            all_days = self.days.lower().split(',')
            if now.format('dd').lower() not in all_days:
                return False

        start_dt = pendulum.parse(self.start_time, tz=self.timezone)\
            .replace(year=now.year, month=now.month, day=now.day)

        end_dt = pendulum.parse(self.end_time, tz=self.timezone)\
            .replace(year=now.year, month=now.month, day=now.day)

        return end_dt >= now >= start_dt


class CapacityDistributionStrategy(enum.Enum):
    spread = 'spread'
    binpack = 'binpack'

    def __repr__(self):
        return repr(self.value)


class CapacityConfiguration(base.Base):

    __tablename__ = 'capacity_config'

    # columns
    id = base.id_builder.build()
    weight_parameter = Column(String(32), nullable=False, unique=True, comment='Parameter name which defines the requested capacity weight.')
    capacity_parameter = Column(String(32), nullable=False, comment='Parameter name which defines the total capacity allocation.')
    producer_parameter = Column(String(32), nullable=False, comment='Parameter name which defines the capacity producers.')
    distribution_strategy = Column(
        Enum(CapacityDistributionStrategy),
        nullable=False,
        default=CapacityDistributionStrategy.binpack,
        comment='Binpack (fill vertically) or spread (fill horizontally)'
    )

    # associations
    producers: List['CapacityProducer'] = relationship(
        'CapacityProducer',
        cascade='save-update, merge, delete, delete-orphan',
        back_populates='configuration'
    )


class CapacityProducer(base.Base):

    __tablename__ = 'capacity_producer'

    # columns
    id = base.id_builder.build()
    configuration_id = Column(base.id_builder.type, ForeignKey('capacity_config.id'), nullable=False)
    name = Column(String(128), nullable=False, comment='The human-readable name of the producer.')
    internal_name = Column(String(128), nullable=False, comment='The internal name of the producer.')
    capacity = Column(Integer, nullable=False, comment='The total capacity of the consumer.')

    # associations
    configuration: 'CapacityConfiguration' = relationship('CapacityConfiguration', back_populates='producers')



import celery
import riberry


class RiberryTask(celery.Task):

    def signature(self, args=None, *starargs, **starkwargs):
        sig = celery.Task.signature(self, args, *starargs, **starkwargs)
        context = riberry.app.current_context
        stream = context.flow.scoped_stream or context.current.stream

        if stream:
            sig['kwargs']['__rib_stream'] = stream
            if '__rib_step' not in sig['kwargs']:
                sig['kwargs']['__rib_step'] = self.name

        return sig



import json
import os
import uuid

import yaml
from sqlalchemy.inspection import inspect
from sqlalchemy.orm.exc import NoResultFound

from riberry import model, services, policy
from riberry.util.common import variable_substitution


class Loader(yaml.SafeLoader):
    """ https://stackoverflow.com/a/9577670 """

    def __init__(self, stream):
        self._root = os.path.split(stream.name)[0]
        super(Loader, self).__init__(stream)

    def include(self, node):
        filename = variable_substitution(os.path.join(self._root, self.construct_scalar(node)))

        with open(filename, 'r') as f:
            if filename.endswith('.yaml') or filename.endswith('.yml'):
                return yaml.load(f, Loader)
            else:
                return f.read()


Loader.add_constructor('!include', Loader.include)


def collection_diff(obj, collection_name, loader):
    current_collection = set(getattr(obj, collection_name))
    new_collection = set(loader())
    for stale in current_collection - new_collection:
        model.conn.delete(stale)
    setattr(obj, collection_name, list(new_collection))
    return obj


def model_diff(obj):
    object_info = inspect(obj)
    return {
        name: attr.history for name, attr in object_info.attrs.items()
        if attr.history.has_changes()
    }


def session_diff():
    diff = {}
    for obj in sorted(model.conn.dirty | model.conn.new | model.conn.deleted, key=lambda o: str(o)):
        type_ = 'Modified' if obj in model.conn.dirty else 'Added' if obj in model.conn.new else 'Deleted' if obj in model.conn.deleted else 'REF'
        diff[obj] = type_, model_diff(obj)
    return diff


def import_applications(applications, restrict=None):
    existing_apps = {a.internal_name: a for a in model.application.Application.query().all()}

    if not restrict:
        for stale in set(existing_apps) - set(applications):
            model.conn.delete(existing_apps[stale])

    apps = []
    for application, properties in applications.items():
        if restrict and application not in restrict:
            continue
        app = import_application(internal_name=application, attributes=properties)
        apps.append(app)
    return apps


def import_application(internal_name, attributes):
    try:
        app = services.application.application_by_internal_name(internal_name=internal_name)
        app = services.application.update_application(app, attributes)
    except NoResultFound:
        app = services.application.create_application(
            internal_name=internal_name,
            name=attributes.get('name'),
            description=attributes.get('description'),
            type=attributes.get('type'),
            document=None
        )

    if attributes.get('document'):
        if not app.document:
            app.document = model.misc.Document()
            model.conn.add(app.document)
        app.document.content = attributes['document'].encode()
    else:
        if app.document:
            app.document = None

    import_instances(app, attributes.get('instances') or {})
    import_forms(app, attributes.get('forms') or {})

    return app


def import_instances(app, instances):
    return collection_diff(
        obj=app,
        collection_name='instances',
        loader=lambda: {
            import_instance(app, name, attrs)
            for name, attrs in instances.items()
        }
    )


def import_forms(app, forms):
    return collection_diff(
        obj=app,
        collection_name='forms',
        loader=lambda: {
            import_form(app, name, attrs)
            for name, attrs in forms.items()
        }
    )


def import_form(app, internal_name, attributes):
    try:
        form = services.form.form_by_internal_name(internal_name=internal_name)
        form = services.form.update_form(form, attributes)
    except NoResultFound:
        instance_internal_name = attributes['instance']
        instances = [instance for instance in app.instances if instance.internal_name == instance_internal_name]

        if not instances:
            raise ValueError(f'Could not find instance {instance_internal_name} for form {internal_name}')

        form = services.form.create_form(
            application=app,
            instance=instances[0],
            name=attributes.get('name'),
            internal_name=internal_name,
            version=attributes.get('version'),
            description=attributes.get('description'),
            input_files=[],
            input_values=[]
        )

    if attributes.get('document'):
        if not form.document:
            form.document = model.misc.Document()
            model.conn.add(form.document)
        form.document.content = attributes['document'].encode()
    else:
        if form.document:
            form.document = None

    import_form_inputs(
        form=form,
        input_files=attributes.get('inputFiles') or {},
        input_values=attributes.get('inputValues') or {},
    )

    return form


def import_input_file_definition(form, internal_name, attributes):
    try:
        definition = services.form.file_definition_by_internal_name(form=form, internal_name=internal_name)
        definition = services.form.update_file_definition(definition, attributes)
    except NoResultFound:
        definition = model.interface.InputFileDefinition(internal_name=internal_name, **attributes)
        model.conn.add(definition)

    return definition


def import_input_value_definition(form, internal_name, attributes):
    mapping = {
        'enumerations': ('allowed_binaries', lambda values: [json.dumps(v).encode() for v in values]),
        'default': ('default_binary', lambda v: json.dumps(v).encode()),
    }

    attributes = dict(
        (mapping[k][0], mapping[k][1](v)) if k in mapping else (k, v)
        for k, v in attributes.items()
    )

    try:
        definition = services.form.value_definition_by_internal_name(form=form, internal_name=internal_name)
        definition = services.form.update_value_definition(definition, attributes)
    except NoResultFound:
        definition = model.interface.InputValueDefinition(internal_name=internal_name, **attributes)
        model.conn.add(definition)

    return definition


def import_form_inputs(form, input_files, input_values):
    collection_diff(
        obj=form,
        collection_name='input_file_definitions',
        loader=lambda: {
            import_input_file_definition(form, name, attrs)
            for name, attrs in input_files.items()
        }
    )

    collection_diff(
        obj=form,
        collection_name='input_value_definitions',
        loader=lambda: {
            import_input_value_definition(form, name, attrs)
            for name, attrs in input_values.items()
        }
    )


def import_instance(app, internal_name, attributes):
    try:
        instance = services.application_instance.application_instance_by_internal_name(internal_name=internal_name)
        instance = services.application_instance.update_application_instance(instance, attributes)
    except NoResultFound:
        instance = services.application_instance.create_application_instance(
            application=app,
            internal_name=internal_name,
            name=attributes.get('name'),
            schedules=[]
        )

    current_schedules = {}
    for schedule in instance.schedules:
        current_schedules[(
            schedule.days,
            schedule.start_time,
            schedule.end_time,
            schedule.timezone,
            schedule.parameter,
            schedule.value,
            schedule.priority,
        )] = schedule

    loaded_schedules = set((
        sched['days'].lower() if isinstance(sched.get('days'), str) else sched.get('days', '*'),
        model.application.ApplicationInstanceSchedule.cleanse_time(
            sched.get('startTime') or model.application.ApplicationInstanceSchedule.start_time.default.arg),
        model.application.ApplicationInstanceSchedule.cleanse_time(
            sched.get('endTime') or model.application.ApplicationInstanceSchedule.end_time.default.arg),
        sched.get('timeZone') or model.application.ApplicationInstanceSchedule.timezone.default.arg,
        sched['parameter'],
        str(sched['value']) if sched['value'] not in ('', None) else None,
        sched.get('priority') or model.application.ApplicationInstanceSchedule.priority.default.arg,
    ) for sched in attributes.get('schedules', []))

    for stale in set(current_schedules) - loaded_schedules:
        model.conn.delete(current_schedules[stale])

    new_schedules = []
    for schedule in loaded_schedules:
        if schedule in current_schedules:
            continue
        days, start_time, end_time, timezone, parameter, value, priority = schedule
        new_schedules.append(
            dict(
                days=days,
                start_time=start_time,
                end_time=end_time,
                timezone=timezone,
                parameter=parameter,
                value=value,
                priority=priority,
            )
        )

    instance.schedules += services.application_instance.create_application_instance_schedules(new_schedules)
    return instance


def import_groups(applications, restrict):
    created_groups = {}
    for application, properties in applications.items():
        if restrict and application not in restrict:
            continue

        for form_internal_name, form_info in properties.get('forms', {}).items():
            groups = form_info.get('groups') or []

            form: model.interface.Form = model.interface.Form.query().filter_by(
                internal_name=form_internal_name,
            ).first()

            if not form:
                print(f'import_groups:: Form {form_internal_name} not found, skipping {application}')
                continue

            associations = {assoc.group.name: assoc for assoc in form.group_associations}
            stale_assocs = set(associations) - set(groups)
            for stale in stale_assocs:
                model.conn.delete(associations[stale])

            for group_name in groups:
                if group_name in associations:
                    continue
                group = model.group.Group.query().filter_by(name=group_name).first()
                if not group:
                    group = created_groups.get(group_name)
                if not group:
                    group = services.auth.create_group(name=group_name)
                    created_groups[group_name] = group

                association = model.group.ResourceGroupAssociation(
                    resource_id=form.id,
                    resource_type=model.misc.ResourceType.form,
                    group=group
                )
                model.conn.add(association)


def _convert_value(value):
    if isinstance(value, list):
        return [_convert_value(v) for v in value]
    if isinstance(value, model.base.Base):
        return dict(
            name=type(value).__name__,
            id=getattr(value, 'id', None)
        )
    if isinstance(value, str) and len(value) > 512:
        return f'(trimmed) {value[:512]}'
    if isinstance(value, bytes):
        return f'bytes (size: {len(value)})'
    import enum
    if isinstance(value, enum.Enum):
        return value.name
    return value


def json_diff(diff):
    output = {'Modified': [], 'Added': [], 'Deleted': []}

    for obj, (diff_type, changes) in diff.items():
        if not changes and diff_type not in ('Added', 'Deleted'):
            continue
        entry = {}
        for k, v in changes.items():
            attr = {}
            if v.deleted:
                attr['Deleted'] = _convert_value(v.deleted[0])
            if v.added and v.added[0]:
                attr['Added'] = _convert_value(v.added[0])
            entry[k] = attr
        obj_dict = _convert_value(obj)
        output[diff_type].append({
            'id': obj_dict['id'],
            'type': obj_dict['name'],
            'attributes': entry
        })

    return output


def import_menu_for_forms(menu):
    for item in model.misc.MenuItem.query().all():
        model.conn.delete(item)

    for item in menu:
        menu_item = import_menu_item(item, menu_type='forms', parent=None)
        model.conn.add(menu_item)


def import_menu_item(item, menu_type, parent=None):
    menu_item = model.misc.MenuItem(parent=parent, menu_type=menu_type)
    if isinstance(item, dict):
        menu_item.key = str(uuid.uuid4())
        menu_item.type = 'branch'
        menu_item.label = list(item.keys())[0]
        menu_item.children = [
            import_menu_item(child, menu_type, menu_item)
            for child in list(item.values())[0]
        ]
    elif isinstance(item, str):
        menu_item.key = item
        menu_item.type = 'leaf'
    return menu_item



def import_config(config, formatter=None, restrict_apps=None):
    applications = config.get('applications') or {}
    import_applications(applications=applications, restrict=restrict_apps)
    if not restrict_apps:
        import_capacities(config.get('capacity-configuration') or {})

    diff = session_diff()
    model.conn.flush()

    import_menu_for_forms(menu=config.get('menues', {}).get('forms', {}))
    import_groups(applications=applications, restrict=restrict_apps)

    for k, v in session_diff().items():
        if k in diff:
            diff[k] = diff[k][0], {**diff[k][1], **v[1]}
        else:
            diff[k] = v

    model.conn.flush()
    return formatter(diff) if formatter else diff


def import_capacities(capacities):
    existing = {c.weight_parameter: c for c in model.application.CapacityConfiguration.query().all()}

    for stale in set(existing) - set(capacities):
        model.conn.delete(existing[stale])

    for weight_name, properties in capacities.items():
        import_capacity(weight_name, properties)


def import_capacity(weight_parameter, properties):
    capacity_config: model.application.CapacityConfiguration = \
        model.application.CapacityConfiguration.query().filter_by(weight_parameter=weight_parameter).first()

    if not capacity_config:
        capacity_config = model.application.CapacityConfiguration(weight_parameter=weight_parameter)

    capacity_config.capacity_parameter = properties['parameters']['capacity']
    capacity_config.producer_parameter = properties['parameters']['producer']

    capacity_config.distribution_strategy = (
        model.application.CapacityDistributionStrategy(properties['strategy'])
        if 'strategy' in properties
        else model.application.CapacityConfiguration.distribution_strategy.default.arg
    )

    new_producers = {p['internalName']: p for p in properties['producers']}

    # Update existing, delete old
    for producer in capacity_config.producers:
        if producer.internal_name not in new_producers:
            model.conn.delete(producer)
        else:
            producer_config = new_producers.pop(producer.internal_name)
            producer.name = producer_config.get('name') or producer.internal_name
            producer.capacity = producer_config['capacity']

    # Add new
    for internal_name, producer_config in new_producers.items():
        capacity_config.producers.append(
            model.application.CapacityProducer(
                internal_name=internal_name,
                name=producer_config.get('name') or internal_name,
                capacity=producer_config['capacity'],
            )
        )

    model.conn.add(capacity_config)


def import_from_file(config_path, dry_run=True, formatter=json_diff, restrict_apps=None):
    with open(config_path) as f:
        config = variable_substitution(yaml.load(f, Loader) or {})

    with model.conn.no_autoflush, policy.context.disabled_scope():
        output = import_config(config, formatter=formatter, restrict_apps=restrict_apps)

    if dry_run:
        model.conn.rollback()
    else:
        model.conn.commit()

    return output



from riberry import policy, model
from riberry.policy import AttributeContext


class RootPolicySet(policy.PolicySet):

    def target_clause(self, context: AttributeContext) -> bool:
        return True

    def condition(self, context: AttributeContext) -> bool:
        return context.subject is not None


class UserPolicySet(policy.PolicySet):

    def target_clause(self, context: AttributeContext) -> bool:
        return True

    def condition(self, context: AttributeContext) -> bool:
        return True


class ApplicationPolicySet(policy.PolicySet):

    def target_clause(self, context: AttributeContext) -> bool:
        return isinstance(context.resource, model.application.Application)

    def condition(self, context: AttributeContext) -> bool:
        return True


class ApplicationViewPolicy(policy.PolicySet):

    def target_clause(self, context: AttributeContext) -> bool:
        return context.action == 'view'

    def condition(self, context: AttributeContext) -> bool:
        return True


class ViewApplicationUsingFormRelationshipRule(policy.Rule):

    def target_clause(self, context: AttributeContext) -> bool:
        return True

    def condition(self, context: AttributeContext) -> bool:
        user: model.auth.User = context.subject
        application: model.application.Application = context.resource

        for instance in application.instances:
            for form in instance.forms:
                if set(user.groups) & set(form.groups):
                    return True
        return False


class GenericCreatePolicy(policy.PolicySet):

    def target_clause(self, context: AttributeContext) -> bool:
        return context.action == 'create'

    def condition(self, context: AttributeContext) -> bool:
        return True


class RejectCreateRule(policy.PolicySet):

    def target_clause(self, context: AttributeContext) -> bool:
        return True

    def condition(self, context: AttributeContext) -> bool:
        return False


class ApplicationInstancePolicySet(policy.PolicySet):

    def target_clause(self, context: AttributeContext) -> bool:
        return isinstance(context.resource, model.application.ApplicationInstance)

    def condition(self, context: AttributeContext) -> bool:
        return True


class ApplicationInstanceViewPolicy(policy.Policy):

    def target_clause(self, context: AttributeContext) -> bool:
        return context.action == 'view'

    def condition(self, context: AttributeContext) -> bool:
        return True


class ViewApplicationInstanceUsingFormRelationshipRule(policy.Rule):

    def target_clause(self, context: AttributeContext) -> bool:
        return True

    def condition(self, context: AttributeContext) -> bool:
        user: model.auth.User = context.subject
        instance: model.application.ApplicationInstance = context.resource

        for form in instance.forms:
            if set(user.groups) & set(form.groups):
                return True
        return False


class ApplicationFormPolicySet(policy.PolicySet):

    def target_clause(self, context: AttributeContext) -> bool:
        return isinstance(context.resource, model.interface.Form)

    def condition(self, context: AttributeContext) -> bool:
        return True


class ApplicationFormViewPolicy(policy.Policy):

    def target_clause(self, context: AttributeContext) -> bool:
        return context.action == 'view'

    def condition(self, context: AttributeContext) -> bool:
        return True


class ViewFormRule(policy.Rule):

    def target_clause(self, context: AttributeContext) -> bool:
        return True

    def condition(self, context: AttributeContext) -> bool:
        user: model.auth.User = context.subject
        form: model.interface.Form = context.resource

        if set(user.groups) & set(form.groups):
            return True
        return False


class JobPolicySet(policy.PolicySet):

    def target_clause(self, context: AttributeContext) -> bool:
        return isinstance(context.resource, model.job.Job)

    def condition(self, context: AttributeContext) -> bool:
        return True


class JobViewPolicy(policy.Policy):

    def target_clause(self, context: AttributeContext) -> bool:
        return context.action == 'view'

    def condition(self, context: AttributeContext) -> bool:
        return True


class ViewJobUsingFormRelationshipRule(policy.Policy):

    def target_clause(self, context: AttributeContext) -> bool:
        return True

    def condition(self, context: AttributeContext) -> bool:
        user: model.auth.User = context.subject
        job: model.job.Job = context.resource

        if set(user.groups) & set(job.form.groups):
            return True
        return False


class JobExecutionPolicySet(policy.PolicySet):

    def target_clause(self, context: AttributeContext) -> bool:
        return isinstance(context.resource, model.job.JobExecution)

    def condition(self, context: AttributeContext) -> bool:
        return True


class JobExecutionViewPolicy(policy.Policy):

    def target_clause(self, context: AttributeContext) -> bool:
        return context.action == 'view'

    def condition(self, context: AttributeContext) -> bool:
        return True


class ViewJobExecutionUsingFormRelationshipRule(policy.Policy):

    def target_clause(self, context: AttributeContext) -> bool:
        return True

    def condition(self, context: AttributeContext) -> bool:
        user: model.auth.User = context.subject
        execution: model.job.JobExecution = context.resource

        if set(user.groups) & set(execution.job.form.groups):
            return True
        return False


class JobExecutionArtifactPolicySet(policy.PolicySet):

    def target_clause(self, context: AttributeContext) -> bool:
        return isinstance(context.resource, model.job.JobExecutionArtifact)

    def condition(self, context: AttributeContext) -> bool:
        return True


class JobExecutionArtifactViewPolicy(policy.Policy):

    def target_clause(self, context: AttributeContext) -> bool:
        return context.action == 'view'

    def condition(self, context: AttributeContext) -> bool:
        return True


class ViewJobExecutionArtifactUsingFormRelationshipRule(policy.Policy):

    def target_clause(self, context: AttributeContext) -> bool:
        return True

    def condition(self, context: AttributeContext) -> bool:
        user: model.auth.User = context.subject
        execution: model.job.JobExecutionArtifact = context.resource

        if set(user.groups) & set(execution.job_execution.job.form.groups):
            return True
        return False


default_policies = policy.AuthorizationEngine(
    'default',
    RootPolicySet(
        UserPolicySet(
            ApplicationPolicySet(
                ApplicationViewPolicy(
                    ViewApplicationUsingFormRelationshipRule()
                ),
                GenericCreatePolicy(
                    RejectCreateRule()
                )
            ),
            ApplicationInstancePolicySet(
                ApplicationInstanceViewPolicy(
                    ViewApplicationInstanceUsingFormRelationshipRule()
                ),
                GenericCreatePolicy(
                    RejectCreateRule()
                )
            ),
            ApplicationFormPolicySet(
                ApplicationFormViewPolicy(
                    ViewFormRule()
                ),
                GenericCreatePolicy(
                    RejectCreateRule()
                )
            ),
            JobPolicySet(
                JobViewPolicy(
                    ViewJobUsingFormRelationshipRule()
                )
            ),
            JobExecutionPolicySet(
                JobExecutionViewPolicy(
                    ViewJobExecutionUsingFormRelationshipRule()
                )
            ),
            JobExecutionArtifactPolicySet(
                JobExecutionArtifactViewPolicy(
                    ViewJobExecutionArtifactUsingFormRelationshipRule()
                )
            )
        )
    )
)


import enum
from collections import defaultdict
from functools import partial

import riberry


class EventRegistryTypes(enum.Enum):

    on_completion = 'on_completion'
    on_data_updated = 'on_data_update'
    on_report_refresh = 'on_report_refresh'
    on_external_result_received = 'on_external_result_received'


class EventRegistry:

    types: EventRegistryTypes = EventRegistryTypes

    def __init__(self, context):
        self.context: riberry.app.context.Context = context
        self._registrations = defaultdict(set)

    @staticmethod
    def _make_key(func, **key):
        if not key and func:
            return tuple([
                ('key', riberry.app.util.misc.function_path(func))
            ])
        return tuple(sorted(key.items()))

    def register(self, event_type: EventRegistryTypes, **key):
        def inner(func):
            formatted_key = self._make_key(func, **key)
            self._registrations[(event_type, formatted_key)].add(func)
            return func
        return inner

    def get(self, event_type: EventRegistryTypes, **key):
        key = self._make_key(func=None, **key)
        return self._registrations[(event_type, key)]

    def call(self, event_type: EventRegistryTypes, args=None, kwargs=None, **key):
        functions = self.get(event_type=event_type, **key)
        return [function(*args or (), **kwargs or {}) for function in functions]


class EventRegistryHelper:

    def __init__(self, context):
        self.context: riberry.app.context.Context = context

    @property
    def _register(self):
        return self.context.event_registry.register

    def execution_failed(self, func):
        key = riberry.app.util.misc.internal_data_key(key='once.execution_failed')
        func = partial(self.context.data.execute_once, key=key, func=func)
        return self._register(event_type=EventRegistryTypes.on_completion, status='FAILURE')(func)

    def execution_succeeded(self, func):
        key = riberry.app.util.misc.internal_data_key(key='once.execution_succeeded')
        func = partial(self.context.data.execute_once, key=key, func=func)
        return self._register(event_type=EventRegistryTypes.on_completion, status='SUCCESS')(func)

    def data_updated(self, name):
        return self._register(event_type=EventRegistryTypes.on_data_updated, data_name=name)

    def external_result_received(self, func):
        return self._register(event_type=EventRegistryTypes.on_external_result_received)(func)

    def report_refresh(self, report, bindings, renderer=None):
        def inner(func):
            def refresh():
                self.context.report.update(report=report, body=func(), renderer=renderer)
            self._register(event_type=EventRegistryTypes.on_report_refresh, report=report)(refresh)
            for binding in bindings:
                self._register(event_type=EventRegistryTypes.on_data_updated, data_name=binding)(refresh)
            return func
        return inner



from celery import signals

import riberry
from ..util.events import create_event


__start_stream_cache = {}


@signals.worker_process_init.connect
def worker_process_init(*args, **kwargs):
    riberry.model.conn.raw_engine.dispose()
    riberry.model.conn.remove()


@signals.celeryd_after_setup.connect
def celeryd_after_setup(*args, **kwargs):
    riberry.model.conn.raw_engine.dispose()
    riberry.model.conn.remove()


@signals.before_task_publish.connect
def before_task_publish(sender, headers, body, **_):
    try:
        root_id = riberry.app.current_context.current.root_id
    except:
        return

    args, kwargs, *_ = body
    task_id = headers['id']

    if '__rib_stream' in kwargs:
        create_event(
            name='stream',
            root_id=root_id,
            task_id=task_id,
            data={
                'stream': str(kwargs['__rib_stream']),
                'state': 'QUEUED'
            }
        )

    if '__rib_step' in kwargs:
        stream, step = kwargs['__rib_stream'], kwargs['__rib_step']
        create_event(
            name='step',
            root_id=root_id,
            task_id=task_id,
            data={
                'stream': str(stream),
                'step': str(step),
                'state': 'QUEUED'
            }
        )


def task_prerun(context, props):
    task_id = context.current.task_id
    root_id = context.current.root_id

    stream = context.current.stream
    step = context.current.step

    if not stream:
        return

    if 'stream_start' in props:
        key = root_id, stream
        if key not in __start_stream_cache:
            __start_stream_cache[key] = None
            if len(__start_stream_cache) > 5000:
                evict_key = next(iter(__start_stream_cache))
                __start_stream_cache.pop(evict_key)
            create_event(
                name='stream',
                root_id=root_id,
                task_id=task_id,
                data={
                    'stream': str(stream),
                    'state': 'ACTIVE'
                }
            )

    if step:
        create_event(
            name='step',
            root_id=root_id,
            task_id=task_id,
            data={
                'stream': str(stream),
                'step': str(step),
                'state': 'ACTIVE'
            }
        )


def task_postrun(context, props, state):
    task_id = context.current.task_id
    root_id = context.current.root_id

    stream = context.current.stream
    step = context.current.step

    if not stream:
        return

    if 'stream_start' in props and state in ('RETRY', 'FAILURE'):
        create_event(
            name='stream',
            root_id=root_id,
            task_id=task_id,
            data={
                'stream': str(stream),
                'state': state,
            }
        )

    if 'stream_end' in props:
        create_event(
            name='stream',
            root_id=root_id,
            task_id=task_id,
            data={
                'stream': str(stream),
                'state': 'SUCCESS' if state == 'IGNORED' else state or 'FAILURE',
            }
        )

    if step:
        create_event(
            name='step',
            root_id=root_id,
            task_id=task_id,
            data={
                'stream': str(stream),
                'step': str(step),
                'state': 'SUCCESS' if state == 'IGNORED' else state or 'FAILURE'
            }
        )



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['riberry',
 'riberry.app',
 'riberry.app.actions',
 'riberry.app.addons',
 'riberry.app.backends',
 'riberry.app.backends.impl',
 'riberry.app.backends.impl.celery',
 'riberry.app.backends.impl.celery.addons',
 'riberry.app.backends.impl.celery.addons.capacity',
 'riberry.app.context',
 'riberry.app.misc',
 'riberry.app.util',
 'riberry.celery',
 'riberry.celery.background',
 'riberry.celery.background.events',
 'riberry.celery.client',
 'riberry.celery.client.dynamic',
 'riberry.celery.client.scale',
 'riberry.model',
 'riberry.model.application',
 'riberry.model.auth',
 'riberry.model.group',
 'riberry.model.interface',
 'riberry.model.job',
 'riberry.model.misc',
 'riberry.plugins',
 'riberry.plugins.defaults',
 'riberry.policy',
 'riberry.services',
 'riberry.util']

package_data = \
{'': ['*']}

install_requires = \
['celery[redis]>=4.3,<5.0',
 'click>=7.0,<8.0',
 'croniter>=0.3.29,<0.4.0',
 'pendulum>=2.0,<3.0',
 'pyjwt>=1.7,<2.0',
 'pyyaml>=5.1,<6.0',
 'redis>=3.2,<4.0',
 'sqlalchemy>=1.3,<2.0',
 'toml>=0.10.0,<0.11.0']

setup_kwargs = {
    'name': 'riberry',
    'version': '0.9.0',
    'description': 'Python-driven job management system',
    'long_description': '# Riberry\n\n[![CircleCI](https://circleci.com/gh/srafehi/riberry/tree/master.svg?style=svg)](https://circleci.com/gh/srafehi/riberry/tree/master)\n\nUnder active development\n\n[Documentation](https://riberry.app)',
    'author': 'Shady Rafehi',
    'author_email': 'shadyrafehi@gmail.com',
    'url': 'https://github.com/srafehi/riberry',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import json
from functools import partial

import riberry


class Report:

    def __init__(self, context):
        self.context: riberry.app.context.Context = context

    def model(self, name):
        job_execution = self.context.current.job_execution
        report: riberry.model.job.JobExecutionReport = riberry.model.job.JobExecutionReport.query().filter_by(
            job_execution=job_execution, name=name,
        ).first()

        if not report:
            self.context.data.execute_once(
                key=riberry.app.util.misc.internal_data_key(f'once.create_report.{name}'),
                func=partial(self._create_report, name=name, job_execution=job_execution)
            )
            return self.model(name=name)

        return report

    @staticmethod
    def _create_report(name, job_execution):
        report = riberry.model.job.JobExecutionReport(job_execution=job_execution, name=name)
        riberry.model.conn.add(report)
        riberry.model.conn.commit()

    def mark_for_refresh(self, name):
        report = self.model(name=name)
        report.marked_for_refresh = True
        riberry.model.conn.commit()

    def update(self, report, body, renderer=None):
        model = self.model(name=report)
        model.marked_for_refresh = False
        model.renderer = renderer or model.renderer
        model.report = json.dumps(body).encode()
        riberry.model.conn.commit()



import types

import celery


def patch_send_task(instance: celery.Celery):
    send_task_original = instance.send_task
    def send_task(self, *args, **kwargs):
        riberry_properties = {}
        if kwargs.get('kwargs'):
            for key, value in kwargs['kwargs'].items():
                if key.startswith('__rib_'):
                    riberry_properties[key.replace('__rib_', '', 1)] = value

        return send_task_original(*args, **kwargs)
    instance.send_task = types.MethodType(send_task, instance)




import hashlib
import os

from riberry.model import auth
from riberry.plugins.interfaces import AuthenticationProvider


def make_hash(text) -> bytes:
    hash_ = hashlib.sha512()
    hash_.update(text)
    return hash_.hexdigest().encode()


def salted_hash(plain: bytes, salt: bytes):
    plain_hash: bytes = make_hash(plain)
    return make_hash(plain_hash + salt)


def check_password(input_password: bytes, hashed_password: bytes):
    size = len(hashed_password) // 2
    password_hash, salt = hashed_password[:size], hashed_password[size:]
    input_hash = salted_hash(input_password, salt) + salt
    return hashed_password == input_hash


def hash_password(password: bytes) -> bytes:
    salt = make_hash(os.urandom(1024))
    return salted_hash(password, salt) + salt


class DefaultAuthenticationProvider(AuthenticationProvider):

    @classmethod
    def name(cls) -> str:
        return 'default'

    def secure_password(self, password: str) -> bytes:
        return hash_password(password=(password or '').encode())

    def authenticate(self, username: str, password: str) -> bool:
        user = auth.User.query().filter_by(username=username).first()
        if user:
            return check_password(input_password=(password or '').encode(), hashed_password=user.password.encode())
        else:
            return False



import inspect
from typing import Optional

import riberry


class ExternalTask:

    def __init__(self, context):
        self.context: riberry.app.context.Context = context

    def create(
            self,
            name: str = None,
            type: str = 'external',
            external_task_id: Optional[str] = None,
            input_data: Optional[bytes] = None
    ):
        external_task = riberry.app.actions.external_task.create_external_task(
            job_execution=self.context.current.job_execution,
            name=name,
            type=type,
            external_task_id=external_task_id,
            input_data=input_data,
        )
        return ExternalTaskCreationResult(context=self.context, instance=external_task)

    def create_receiver_task(self, external_task_id, validator):
        if inspect.isfunction(validator):
            validator = riberry.app.util.misc.function_path(validator)

        return self.context.current.riberry_app.backend.create_receiver_task(
            external_task_id=external_task_id,
            validator=validator,
        )

    @staticmethod
    def mark_as_ready(external_task_id, output_data):
        return riberry.app.actions.external_task.mark_as_ready(
            external_task_id=external_task_id,
            output_data=output_data
        )


class ExternalTaskCreationResult:

    def __init__(self, context, instance):
        self.context: riberry.app.context.Context = context
        self.instance = instance
        self.task_id = instance.task_id

    def create_receiver_task(self, validator=None):
        return self.context.external_task.create_receiver_task(external_task_id=self.task_id, validator=validator)



import json
import mimetypes
from typing import List

from sqlalchemy import Column, String, Boolean, ForeignKey, Integer, Binary, DateTime, desc
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import relationship, deferred

from riberry import model
from riberry.model import base


class Form(base.Base):
    """
    A Form is an interface to creating jobs for a given ApplicationInstance.
    """

    __tablename__ = 'form'
    __reprattrs__ = ['internal_name', 'version']

    # columns
    id = base.id_builder.build()
    instance_id = Column(base.id_builder.type, ForeignKey('app_instance.id'), nullable=False)
    application_id = Column(base.id_builder.type, ForeignKey('application.id'), nullable=False)
    document_id = Column(base.id_builder.type, ForeignKey(column='document.id'))
    name: str = Column(String(64), unique=True, nullable=False, comment='The human-readable name of the form.')
    internal_name: str = Column(String(256), unique=True, nullable=False, comment='The internal name or secondary identifier of the form.')
    description: str = Column(String(256), comment='A brief description of the form\'s purpose.')
    version: int = Column(Integer, nullable=False, default=1, comment='The version of the form.')
    enabled: bool = Column(Boolean, nullable=False, default=True, comment='Whether or not this form is enabled.')

    # associations
    instance: 'model.application.ApplicationInstance' = relationship('ApplicationInstance', back_populates='forms')
    application: 'model.application.Application' = relationship('Application', back_populates='forms')
    schedules: List['FormSchedule'] = relationship('FormSchedule', back_populates='form')
    jobs: List['model.job.Job'] = relationship(
        'Job',
        cascade='save-update, merge, delete, delete-orphan',
        order_by=lambda: desc(model.job.Job.created),
        back_populates='form'
    )
    group_associations: List['model.group.ResourceGroupAssociation'] = model.group.ResourceGroupAssociation.make_relationship(
        resource_id=id,
        resource_type=model.misc.ResourceType.form
    )
    input_value_definitions: List['InputValueDefinition'] = relationship(
        'InputValueDefinition',
        cascade='save-update, merge, delete, delete-orphan',
        order_by=lambda: InputValueDefinition.id.asc(),
        back_populates='form'
    )
    input_file_definitions: List['InputFileDefinition'] = relationship(
        'InputFileDefinition',
        cascade='save-update, merge, delete, delete-orphan',
        order_by=lambda: InputFileDefinition.id.asc(),
        back_populates='form'
    )
    document: 'model.misc.Document' = relationship('Document', cascade='save-update, merge, delete, delete-orphan', single_parent=True)

    # proxies
    groups: List['model.group.Group'] = association_proxy('group_associations', 'group')


class FormSchedule(base.Base):
    __tablename__ = 'sched_form'

    # columns
    id = base.id_builder.build()
    form_id = Column(base.id_builder.type, ForeignKey('form.id'), nullable=False)
    start = Column(DateTime(timezone=True), nullable=False)
    end = Column(DateTime(timezone=True), nullable=False)

    # associations
    form: 'Form' = relationship('Form', back_populates='schedules')


class InputFileDefinition(base.Base):
    """The InputFileDefinition object defines the properties of an input file."""

    __tablename__ = 'input_file_definition'
    __reprattrs__ = ['name', 'type']

    # columns
    id = base.id_builder.build()
    form_id = Column(base.id_builder.type, ForeignKey('form.id'), nullable=False)

    name: str = Column(String(64), nullable=False)
    internal_name: str = Column(String(256), nullable=False)
    description: str = Column(String(128))
    type: str = Column(String(64), nullable=False)
    accept: str = Column(String(256))
    required: bool = Column(Boolean, nullable=False, default=True)

    # associations
    form: 'Form' = relationship('Form', back_populates='input_file_definitions')


class InputValueDefinition(base.Base):
    """The InputFileDefinition object defines the properties of an input value."""

    __tablename__ = 'input_value_definition'
    __reprattrs__ = ['name', 'type']

    # columns
    id = base.id_builder.build()
    form_id = Column(base.id_builder.type, ForeignKey('form.id'), nullable=False)

    name: str = Column(String(64), nullable=False)
    internal_name: str = Column(String(256), nullable=False)
    description: str = Column(String(128))
    type: str = Column(String(64), nullable=False)
    required: bool = Column(Boolean, nullable=False, default=True)
    default_binary = Column('defaults', Binary)

    # associations
    form: 'Form' = relationship('Form', back_populates='input_value_definitions')
    allowed_value_enumerations: List['InputValueEnum'] = relationship(
        'InputValueEnum', cascade='save-update, merge, delete, delete-orphan', back_populates='definition')

    # proxies
    allowed_binaries: List[bytes] = association_proxy(
        'allowed_value_enumerations',
        'value',
        creator=lambda value: InputValueEnum(value=value)
    )

    @property
    def allowed_values(self):
        return [json.loads(v.decode()) for v in self.allowed_binaries]

    @hybrid_property
    def default_value(self):
        return json.loads(self.default_binary.decode()) if self.default_binary else None

    @default_value.setter
    def default_value(self, value):
        self.default_binary = json.dumps(value).encode()


class InputValueEnum(base.Base):
    """The InputValueEnum object defines a valid enumeration for a given InputValueInstance."""

    __tablename__ = 'input_value_enum'
    __reprattrs__ = ['value']

    # columns
    id = base.id_builder.build()
    definition_id = Column(base.id_builder.type, ForeignKey('input_value_definition.id'), nullable=False)
    value = Column(Binary, nullable=False)

    # associations
    definition: 'InputValueDefinition' = relationship(
        'InputValueDefinition', back_populates='allowed_value_enumerations')


class InputValueInstance(base.Base):
    """The InputValueInstance object contains data for a InputValueDefinition and is linked to a Job."""

    __tablename__ = 'input_value_instance'

    # columns
    id = base.id_builder.build()
    job_id = Column(base.id_builder.type, ForeignKey('job.id'), nullable=False)
    name: str = Column(String(256), nullable=False)
    internal_name: str = Column(String(256), nullable=False)
    raw_value: bytes = Column('value', Binary)

    # associations
    job: 'model.job.Job' = relationship('Job', back_populates='values')

    @property
    def definition(self):
        return self.job.form

    @hybrid_property
    def value(self):
        return json.loads(self.raw_value.decode()) if self.raw_value else None

    @value.setter
    def value(self, value):
        self.raw_value = json.dumps(value).encode()


class InputFileInstance(base.Base):
    """The InputFileInstance object contains data for a InputFileDefinition and is linked to a Job."""

    __tablename__ = 'input_file_instance'
    __reprattrs__ = ['filename', 'size']

    # columns
    id = base.id_builder.build()
    job_id = Column(base.id_builder.type, ForeignKey('job.id'), nullable=False)
    name: str = Column(String(256), nullable=False)
    internal_name: str = Column(String(256), nullable=False)
    filename: str = Column(String(512), nullable=False)
    size: int = Column(Integer, nullable=False)
    binary: bytes = deferred(Column(Binary))

    # associations
    job: 'model.job.Job' = relationship('Job', back_populates='files')

    @property
    def content_type(self):
        if self.filename.endswith('.log'):  # quick fix for missing .log type on unix systems
            return 'text/plain'

        return mimetypes.guess_type(self.filename)[0]

    @property
    def content_encoding(self):
        return mimetypes.guess_type(self.filename)[1]



import riberry
from celery import shared_task


@shared_task(name='riberry.core.execution_complete', bind=True, ignore_result=True)
def execution_complete(task, status, stream):
    with riberry.model.conn:
        return riberry.app.actions.executions.execution_complete(
            task_id=task.request.id, root_id=task.request.root_id, status=status, stream=stream
        )



import importlib

from riberry import model
from riberry.celery.background import capacity_config
from riberry.celery.background.events import events
from . import app


@app.task(ignore_result=True)
def process_events(event_limit=None):
    events.process(event_limit)


@app.task(ignore_result=True)
def job_schedules():
    with model.conn:
        for schedule in model.job.JobSchedule.query().filter_by(enabled=True).all():
            schedule.run()

        model.conn.commit()


@app.task(ignore_result=True)
def update_capacity_parameters():
    with model.conn:
        for capacity_configuration in model.application.CapacityConfiguration.query().all():
            producers = [
                capacity_config.CapacityProducer(producer.internal_name, producer.capacity) for producer in
                capacity_configuration.producers
            ]

            capacity_config.update_instance_capacities(
                producers=producers,
                weight_parameter=capacity_configuration.weight_parameter,
                capacity_parameter=capacity_configuration.capacity_parameter,
                producer_parameter=capacity_configuration.producer_parameter,
                distribution_strategy=capacity_configuration.distribution_strategy,
            )

        model.conn.commit()


@app.task(ignore_result=True)
def custom_task(func_path):
    module_path, func_name = func_path.split(':')
    module = importlib.import_module(module_path)
    func = getattr(module, func_name)
    func()



from collections import Counter, defaultdict
from enum import Enum
from typing import Optional, Dict, List, Set

import itertools
from celery.utils.log import logger

from riberry import model


class ConsumerStatus(Enum):
    active = 'active'
    inactive = 'inactive'


class CapacityConsumer:

    def __init__(self, name: str, status: ConsumerStatus, requested_capacity: Optional[int] = None):
        self.name = name
        self.status = status
        self.requested_capacity = requested_capacity

    def __repr__(self):
        return f'CapacityConsumer(name={self.name!r})'

    @staticmethod
    def total_capacity(consumers, default=0):
        return sum(c.requested_capacity or default for c in consumers if c.status == ConsumerStatus.active)

    @classmethod
    def distribute(cls, consumers, total_capacity) -> Dict['CapacityConsumer', List[int]]:
        total_requested_capacity = cls.total_capacity(consumers=consumers, default=total_capacity) or total_capacity
        return {
            consumer: (
                [(consumer.requested_capacity or total_capacity) / total_requested_capacity * total_capacity]
                if consumer.status == ConsumerStatus.active else [0]
            )
            for consumer in consumers
        }

    @classmethod
    def from_weight_parameter(cls, parameter_name: str):
        schedules: List[model.application.ApplicationInstanceSchedule] = \
            model.application.ApplicationInstanceSchedule.query().filter_by(parameter=parameter_name).all()
        instances: Set[model.application.ApplicationInstance] = {sched.instance for sched in schedules}
        schedule_values = {
            instance.internal_name: int(instance.active_schedule_value(name=parameter_name, default=0))
            for instance in instances
        }
        execution_count = execution_count_for_instances(instances=instances)

        return [
            CapacityConsumer(
                name=instance.internal_name,
                status=(
                    ConsumerStatus.active
                    if instance.status == 'online' and execution_count[instance.internal_name]
                    else ConsumerStatus.inactive
                ),
                requested_capacity=schedule_values[instance.internal_name]
            ) for instance in instances
        ]


class CapacityProducer:

    def __init__(self, name: str, capacity: int):
        self.name = name
        self.capacity = capacity

    def __repr__(self):
        return f'CapacityProducer(name={self.name!r}, capacity={self.capacity})'

    @staticmethod
    def total_capacity(producers):
        return sum(p.capacity for p in producers)

    @staticmethod
    def producers_name_pool(producers, distribution_strategy: model.application.CapacityDistributionStrategy):
        name_lists = sorted(
            [[producer.name] * producer.capacity for producer in producers],
            key=len,
            reverse=True
        )

        if distribution_strategy == model.application.CapacityDistributionStrategy.spread:
            name_lists = [filter(None, name_list) for name_list in itertools.zip_longest(*name_lists)]

        return list(itertools.chain.from_iterable(name_lists))


def weighted_schedules(parameter_name: str):
    schedules = model.application.ApplicationInstanceSchedule.query().filter_by(parameter=parameter_name).all()
    return schedules


def execution_count_for_instances(instances, states=('ACTIVE', 'READY')):
    execution_count = defaultdict(int)
    job_executions: List[model.job.JobExecution] = model.job.JobExecution.query().filter(
        model.job.JobExecution.status.in_(states)).all()

    for job_execution in job_executions:
        instance = job_execution.job.instance
        if instance in instances:
            execution_count[instance.internal_name] += 1

    return execution_count


def update_instance_schedule(
        instance: model.application.ApplicationInstance,
        capacity: int,
        producer_allocations: Counter,
        allocation_config_name: str,
        capacity_config_name: str
):
    for sched in list(instance.schedules):
        if sched.parameter in (capacity_config_name, allocation_config_name):
            model.conn.delete(sched)

    schedule_capacity = model.application.ApplicationInstanceSchedule(
        instance=instance,
        parameter=capacity_config_name,
        value=str(capacity),
        priority=100,
    )

    schedule_allocation = model.application.ApplicationInstanceSchedule(
        instance=instance,
        parameter=allocation_config_name,
        value=' '.join(f'{k}|{v}' for k, v in sorted(producer_allocations.items())),
        priority=101,
    )

    model.conn.add(schedule_capacity)
    model.conn.add(schedule_allocation)


def update_instance_capacities(
        producers, weight_parameter, capacity_parameter, producer_parameter, distribution_strategy):

    consumers = CapacityConsumer.from_weight_parameter(parameter_name=weight_parameter)

    total_capacity = CapacityProducer.total_capacity(producers=producers)
    capacity_distribution = CapacityConsumer.distribute(consumers=consumers, total_capacity=total_capacity)
    producer_name_pool = CapacityProducer.producers_name_pool(
        producers=producers, distribution_strategy=distribution_strategy)

    logger.info(f'[{weight_parameter}] Total capacity: {total_capacity}')

    for consumer, capacities in sorted(capacity_distribution.items(), key=lambda x: x[0].name):
        raw_capacity = round(sum(capacities))
        allocated_names = producer_name_pool[:raw_capacity]
        producer_name_pool = producer_name_pool[raw_capacity:]

        capacity = len(allocated_names)
        producer_allocations = Counter(allocated_names)

        instance = model.application.ApplicationInstance.query().filter_by(internal_name=consumer.name).one()
        update_instance_schedule(
            instance=instance,
            capacity=capacity,
            producer_allocations=producer_allocations,
            allocation_config_name=producer_parameter,
            capacity_config_name=capacity_parameter,
        )

        allocations_formatted = ', '.join([f'{k}: {v:2}' for k, v in sorted(producer_allocations.items())]) or '-'
        logger.info(
            f'[{weight_parameter}] {consumer.name} -> capacity: {capacity:2}, allocations: [ {allocations_formatted} ]')



import csv
import io
import json
from collections import Mapping
from operator import itemgetter
from typing import Union, AnyStr, Iterator, Any, Dict

import riberry


class InputMappings:

    def __init__(self, context):
        self.values = InputValueMapping(context=context)
        self.files = InputFileMapping(context=context)


class InputMapping(Mapping):

    def __init__(self, context, cls):
        self.context: riberry.app.context.Context = context
        self.cls = cls

    def _get_value(self, instance):
        raise NotImplementedError

    def get(self, item, default=None):
        if isinstance(item, str):
            return super().get(item, default=default)

        return [super(InputMapping, self).get(_, default) for _ in item]

    def __getitem__(self, item: Union[AnyStr, Iterator[AnyStr]]) -> Any:

        query = self.cls.query().filter(self.cls.job == self.context.current.job)

        if isinstance(item, str):
            instance = query.filter_by(internal_name=item).first()
            if not instance:
                raise KeyError(item)
            return self._get_value(instance)
        else:
            instances = query.filter(self.cls.internal_name.in_(item)).all()
            mapping = {instance.internal_name: self._get_value(instance) for instance in instances}
            return tuple(mapping[key] for key in item)

    def __len__(self) -> int:
        return self.cls.query().filter_by(job=self.context.current.job).count()

    def __iter__(self) -> Iterator[AnyStr]:
        instances = riberry.model.conn.query(
            self.cls.internal_name
        ).filter_by(
            job=self.context.current.job
        ).all()

        return map(itemgetter(0), instances)

    @property
    def dict(self) -> Dict[AnyStr, Any]:
        return {
            instance.internal_name: instance.value
            for instance in self.cls.query().filter_by(job=self.context.current.job).all()
        }


class InputValueMapping(InputMapping):

    def __init__(self, context):
        super().__init__(context=context, cls=riberry.model.interface.InputValueInstance)

    def _get_value(self, instance):
        return instance.value


class InputFileMapping(InputMapping):

    def __init__(self, context):
        super().__init__(context=context, cls=riberry.model.interface.InputFileInstance)

    def _get_value(self, instance):
        return InputFileReader(instance)

    def __getitem__(self, item: Union[AnyStr, Iterator[AnyStr]]) -> Union[
        'InputFileReader', Iterator['InputFileReader']]:
        return super().__getitem__(item)


class InputFileReader:
    instance: riberry.model.interface.InputFileInstance

    def __init__(self, file_instance):
        self.instance = file_instance

    @property
    def filename(self):
        return self.instance.filename

    @property
    def size(self):
        return self.instance.size

    def bytes(self) -> bytes:
        return self.instance.binary

    def text(self, *args, **kwargs) -> str:
        return self.bytes().decode(*args, **kwargs)

    def json(self, *args, **kwargs):
        return json.loads(self.bytes(), *args, **kwargs)

    def csv(self, *args, **kwargs):
        reader = csv.DictReader(io.StringIO(self.text()), *args, **kwargs)
        for row in reader:
            yield row

    def __bool__(self):
        return bool(self.instance.binary)
